from django import forms
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from wagtail.admin.widgets import AdminPageChooser

from wagtail_localize.models import Region


class RegionForm(forms.ModelForm):
    required_css_class = 'required'

    class Meta:
        model = Region
        fields = ['name', 'slug', 'languages', 'is_active']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['languages'].queryset = self.fields['languages'].queryset.filter(is_active=True)



from django.conf.urls import include, url

from wagtail.core import hooks

from wagtail_localize.admin.workflow.action_modules import BaseActionModule

from .views import language_code, translate


class SubmitToGoogleTranslateActionModule(BaseActionModule):
    template_name = 'wagtail_localize_google_translate/action_module.html'

    def is_shown(self):
        # Hide if the language is not different between the locales
        source_lang = language_code(self.translation_request.source_locale.language.code)
        target_lang = language_code(self.translation_request.target_locale.language.code)

        return source_lang != target_lang


@hooks.register('wagtail_localize_workflow_register_action_modules')
def wagtail_localize_workflow_register_action_modules():
    return [
        SubmitToGoogleTranslateActionModule,
    ]


@hooks.register('register_admin_urls')
def register_admin_urls():
    urls = [
        url('^translate/(\d+)/$', translate, name='translate'),
    ]

    return [
        url('^localize/engine/googletranslate/', include((urls, 'wagtail_localize_google_translate'), namespace='wagtail_localize_google_translate')),
    ]



from django.apps import AppConfig


class WagtailLocalizeRegionsAdminAppConfig(AppConfig):
    label = 'wagtail_localize_regions_admin'
    name = 'wagtail_localize.admin.regions'
    verbose_name = "Wagtail Localize Regions admin"



import functools

from django.conf import settings
from django.conf.locale import LANG_INFO
from django.utils.translation import check_for_language


@functools.lru_cache()
def get_languages():
    """
    Cache of settings.LANGUAGES in a dictionary for easy lookups by key.
    """
    return dict(settings.LANGUAGES)


# Added in Django 2.1
@functools.lru_cache(maxsize=1000)
def get_supported_language_variant(lang_code, strict=False):
    """
    Return the language code that's listed in supported languages, possibly
    selecting a more generic variant. Raise LookupError if nothing is found.
    If `strict` is False (the default), look for a country-specific variant
    when neither the language code nor its generic variant is found.
    lru_cache should have a maxsize to prevent from memory exhaustion attacks,
    as the provided language codes are taken from the HTTP request. See also
    <https://www.djangoproject.com/weblog/2007/oct/26/security-fix/>.
    """
    if lang_code:
        # If 'fr-ca' is not supported, try special fallback or language-only 'fr'.
        possible_lang_codes = [lang_code]
        try:
            possible_lang_codes.extend(LANG_INFO[lang_code]['fallback'])
        except KeyError:
            pass
        generic_lang_code = lang_code.split('-')[0]
        possible_lang_codes.append(generic_lang_code)
        supported_lang_codes = get_languages()

        for code in possible_lang_codes:
            if code in supported_lang_codes and check_for_language(code):
                return code
        if not strict:
            # if fr-fr is not supported, try fr-ca.
            for supported_code in supported_lang_codes:
                if supported_code.startswith(generic_lang_code + '-'):
                    return supported_code
    raise LookupError(lang_code)



from django.conf import settings
from django.db import models
from django.db.models import Case, Exists, OuterRef, When

from wagtail.core.models import Page, PageRevision
from wagtail_localize.models import Locale


class TranslationRequestQuerySet(models.QuerySet):
    def annotate_completed(self):
        """
        Adds an is_completed flag to each translation request. Which is true if at least
        one page is completed and no pages are in progress
        """
        in_progress_pages = TranslationRequestPage.objects.filter(request=OuterRef('pk')).in_progress()
        completed_pages = TranslationRequestPage.objects.filter(request=OuterRef('pk')).completed()

        return self.annotate(
            has_in_progress_pages=Exists(in_progress_pages),
            has_completed_pages=Exists(completed_pages),
            is_completed=Case(
                When(has_in_progress_pages=False, has_completed_pages=True, then=True),
                default=False,
                output_field=models.BooleanField(),
            )
        )

    def completed(self):
        """
        Filters the queryset to contain only translation requests that have been translated.

        In order for this to be true, at least one of the translation pages must be translated
        and all others must not be in progress.
        """
        return self.annotate_completed().filter(is_completed=True)


class TranslationRequest(models.Model):
    # Protected foreign keys because locales shouldn't be deleted without manually dealing
    # with translation requests first
    source_locale = models.ForeignKey(Locale, on_delete=models.PROTECT, related_name='+')
    target_locale = models.ForeignKey(Locale, on_delete=models.PROTECT, related_name='+')

    # A foreign key to the page in destination tree where the translated pages will be created
    # If this is deleted before this translation request is complete, any remaining translations will be cancelled
    target_root = models.ForeignKey(Page, null=True, on_delete=models.SET_NULL, related_name='+')

    created_at = models.DateTimeField()
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name='translation_requests_created')

    objects = TranslationRequestQuerySet.as_manager()

    @property
    def is_cancelled(self):
        return self.target_root_id is None

    def get_status(self):
        # Are there any pages awaiting translation?
        if self.pages.in_progress().exists():
            if self.target_root_id is None:
                return "Cancelled (target page deleted)"
            else:
                return "In progress"

        else:
            # Has at least one page been translated successfully?
            if self.pages.completed().exists():
                return "Completed"
            else:
                return "Cancelled (all source pages deleted)"


class TranslationRequestPageQuerySet(models.QuerySet):
    def completed(self):
        """
        Filters the queryset to contain only translation request pages that have been successfully translated.
        """
        return self.filter(is_completed=True)

    def cancelled(self):
        """
        Filters the queryset to only include translation request pages which have been cancelled.

        Translation request pages will be automatically cancelled if their source page is deleted before
        the translation was complete.

        Note: this does not take into account the cancellation status of the translation request as a whole.
        """
        return self.filter(is_completed=False, source_revision__isnull=True)

    def in_progress(self):
        """
        Filters the queryset to include translation request pages that are awaiting translations.

        These include all incomplete translations that haven't been cancelled.

        Note: this does not take into account the cancellation status of the translation request as a whole.
        """
        return self.filter(is_completed=False, source_revision__isnull=False)


class TranslationRequestPage(models.Model):
    request = models.ForeignKey(TranslationRequest, on_delete=models.CASCADE, related_name='pages')

    # source_revision will only be null if the original page or revision has been deleted.
    # in this case, we will treat this translation request as "Cancelled"
    source_revision = models.ForeignKey(PageRevision, null=True, on_delete=models.SET_NULL, related_name='+')

    # When a whole tree is submitted for translation, this allows that tree to be represented in translation requests.
    # When the translated pages are created, this is used to structure those pages in the same way as the source tree.
    # If this is null, the translated page must be created directly under request.target_root. All transation request
    # must have at least one page that has parent=null.
    parent = models.ForeignKey('self', on_delete=models.PROTECT, null=True, related_name='child_pages')

    is_completed = models.BooleanField(default=False)

    # The revision that was created when this change was made
    # This is nullable because pages can be deleted later on
    completed_revision = models.ForeignKey(PageRevision, on_delete=models.SET_NULL, null=True, related_name='+')

    objects = TranslationRequestPageQuerySet.as_manager()

    @property
    def is_cancelled(self):
        # TODO: If this page's parent was cancelled, this should probably be marked as cancelled as well.
        return self.source_revision_id is None and not self.is_completed

    @property
    def source_page(self):
        return self.source_revision.page

    @property
    def previous_request(self):
        return TranslationRequestPage.objects.filter(source_revision__page=self.source_page, request__target_locale=self.request.target_locale, id__lt=self.id).order_by('-request__created_at').first()

    def get_translation(self):
        return self.source_page.specific.get_translation(self.request.target_locale)



import functools

from wagtail.admin.edit_handlers import ObjectList, extract_panel_definitions_from_model_class


REGION_COMPONENTS = []


def get_region_components():
    return REGION_COMPONENTS


def register_region_component(model):
    if model not in REGION_COMPONENTS:
        REGION_COMPONENTS.append(model)
        REGION_COMPONENTS.sort(key=lambda x: x._meta.verbose_name)

    return model


@functools.lru_cache()
def get_region_component_edit_handler(model):
    if hasattr(model, 'edit_handler'):
        # use the edit handler specified on the class
        return model.edit_handler
    else:
        panels = extract_panel_definitions_from_model_class(model, exclude=['region'])
        return ObjectList(panels)



from django.apps import AppConfig


class WagtailLocalizePOFileAppConfig(AppConfig):
    label = 'wagtail_localize_pofile'
    name = 'wagtail_localize.translation_engines.pofile'
    verbose_name = "Wagtail Localize PO file translation engine"



from collections import defaultdict

from django.conf import settings
from django.shortcuts import get_object_or_404

from wagtail.admin.modal_workflow import render_modal_workflow
from wagtail.core.models import Page

from wagtail_localize.models import Region


def translations_list(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific
    regions = list(Region.objects.filter(is_active=True))

    # Get translations by region
    translations_by_region_unsorted = defaultdict(list)
    for translation in page.get_translations(inclusive=True).select_related('locale'):
        translations_by_region_unsorted[translation.locale.region_id].append(translation)

    # Now make them a list using region default ordering
    translations_by_region = [
        (region, translations_by_region_unsorted[region.id])
        for region in regions
        if region.id in translations_by_region_unsorted
    ]

    return render_modal_workflow(request, 'wagtail_localize_language_switch/translation_list.html', None, {
        'page': page,
        'multi_region': len(regions) > 1,
        'translations_by_region': translations_by_region,
    })



from django import forms
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from wagtail.admin.widgets import AdminPageChooser

from wagtail_localize.models import Region
from .models import Site, SiteLanguage


class SiteForm(forms.ModelForm):
    required_css_class = "required"

    class Meta:
        model = Site
        fields = ['hostname', 'port', 'site_name', 'is_default_site']


class SiteLanguageForm(forms.ModelForm):

    class Meta:
        model = SiteLanguage
        fields = ['region', 'language', 'root_page', 'is_active']


SiteLanguageFormSetBase = forms.inlineformset_factory(Site, SiteLanguage, form=SiteLanguageForm, extra=0)

class SiteLanguageFormSet(SiteLanguageFormSetBase):
    minimum_forms = 1
    minimum_forms_message = _("Please specify at least one language for this site.")

    @cached_property
    def single_region(self):
        return Region.objects.count() == 1

    def add_fields(self, form, *args, **kwargs):
        super().add_fields(form, *args, **kwargs)

        # Hide delete field
        form.fields['DELETE'].widget = forms.HiddenInput()

        # Use page chooser for root page
        form.fields['root_page'].widget = AdminPageChooser()

        # Hide region field if there is only one
        if self.single_region:
            form.fields['region'].widget = forms.HiddenInput()



from django import template

from wagtail_localize.models import TranslatablePageMixin


register = template.Library()


@register.inclusion_tag('wagtail_localize_language_switch/_language_switch.html', takes_context=True)
def admin_language_switch(context):
    page = context['page']
    is_translatable = True

    if isinstance(page, TranslatablePageMixin):
        language_count = page.get_translations().count()
    else:
        is_translatable = False
        language_count = 0

    return {
        'page': page,
        'is_translatable': is_translatable,
        'language_count': language_count,
    }



def find_available_slug(parent, requested_slug):
    existing_slugs = set(parent.get_children().filter(slug__startswith=requested_slug).values_list('slug', flat=True))
    slug = requested_slug
    number = 1

    while slug in existing_slugs:
        slug = requested_slug + '-' + str(number)
        number += 1

    return slug



from collections import defaultdict

from django.db import models

from wagtail.core import blocks
from wagtail.core.fields import RichTextField, StreamField
from wagtail.core.rich_text import RichText
from wagtail.images.blocks import ImageChooserBlock
from wagtail.snippets.blocks import SnippetChooserBlock

from wagtail_localize.models import TranslatableMixin
from wagtail_localize.segments import TemplateValue

from .html import restore_html_segments


def organise_template_segments(segments):
    template = None
    segments_by_position = {}

    for segment in segments:
        if isinstance(segment, TemplateValue):
            template = segment
        else:
            segments_by_position[int(segment.path)] = segment

    texts = []
    for position in range(template.segment_count):
        texts.append(segments_by_position[position].text)

    return template.format, template.template, texts


def handle_related_object(related_object, src_locale, tgt_locale, segments):
    if related_object is None or not isinstance(related_object, TranslatableMixin):
        return related_object

    # Note: when called from streamfield, we may be given the translated object
    related_original = related_object.get_translation(src_locale)
    related_translated = related_object.get_translation_or_none(tgt_locale)

    if related_translated is None:
        # Create translated version by copying the original version
        related_translated = related_original.copy_for_translation(tgt_locale)

    ingest_segments(related_original, related_translated, src_locale, tgt_locale, segments)
    related_translated.save()
    return related_translated


class StreamFieldSegmentsWriter:
    def __init__(self, field, src_locale, tgt_locale):
        self.field = field
        self.src_locale = src_locale
        self.tgt_locale = tgt_locale

    def handle_block(self, block_type, block_value, segments):
        if hasattr(block_type, 'restore_translated_segments'):
            return block_type.restore_translated_segments(block_value, segments)

        elif isinstance(block_type, blocks.CharBlock):
            return segments[0].text

        elif isinstance(block_type, blocks.RichTextBlock):
            format, template, texts = organise_template_segments(segments)
            assert format == 'html'
            return RichText(restore_html_segments(template, texts))

        elif isinstance(block_type, (ImageChooserBlock, SnippetChooserBlock)):
            return self.handle_related_object_block(block_value, segments)

        elif isinstance(block_type, blocks.StructBlock):
            return self.handle_struct_block(block_value, segments)

        elif isinstance(block_type, blocks.ListBlock):
            return self.handle_list_block(block_value, segments)

        else:
            raise Exception("Unrecognised StreamField block type '{}'. Have you implemented restore_translated_segments() on this class?".format(block_type.__class__.__name__))

    def handle_related_object_block(self, related_object, segments):
        return handle_related_object(related_object, self.src_locale, self.tgt_locale, segments)

    def handle_struct_block(self, struct_block, segments):
        segments_by_field = defaultdict(list)

        for segment in segments:
            field_name, segment = segment.unwrap()
            segments_by_field[field_name].append(segment)

        for field_name in getattr(struct_block.block.meta, 'translated_fields', []):
            segments = segments_by_field[field_name]

            if segments:
                block_type = struct_block.block.child_blocks[field_name]
                block_value = struct_block[field_name]
                struct_block[field_name] = self.handle_block(block_type, block_value, segments)

        return struct_block

    def handle_list_block(self, list_block, segments):
        # TODO
        pass

    def get_stream_block_child_data(self, data, block_uuid):
        for stream_child in data:
            if stream_child.id == block_uuid:
                return stream_child

    def handle_stream_block(self, data, segments):
        segments_by_block = defaultdict(list)

        for segment in segments:
            block_uuid, segment = segment.unwrap()
            segments_by_block[block_uuid].append(segment)

        for block_uuid, segments in segments_by_block.items():
            block = self.get_stream_block_child_data(data, block_uuid)
            block.value = self.handle_block(block.block, block.value, segments)


def ingest_segments(original_obj, translated_obj, src_locale, tgt_locale, segments):
    # Get segments by field name
    segments_by_field_name = defaultdict(list)
    for segment in segments:
        field_name, segment = segment.unwrap()
        segments_by_field_name[field_name].append(segment)

    for field_name, field_segments in segments_by_field_name.items():
        field = translated_obj.__class__._meta.get_field(field_name)

        if hasattr(field, 'restore_translated_segments'):
            value = field.value_from_object(translated_obj)
            new_value = field.restore_translated_segments(value, field_segments)
            setattr(translated_obj, field_name, new_value)

        elif isinstance(field, StreamField):
            data = field.value_from_object(translated_obj)
            StreamFieldSegmentsWriter(field, src_locale, tgt_locale).handle_stream_block(data, field_segments)
            setattr(translated_obj, field_name, data)

        elif isinstance(field, RichTextField):
            format, template, texts = organise_template_segments(field_segments)
            assert format == 'html'
            html = restore_html_segments(template, texts)
            setattr(translated_obj, field_name, RichText(html))

        elif isinstance(field, (models.TextField, models.CharField)):
            setattr(translated_obj, field_name, field_segments[0].text)

        elif isinstance(field, models.ForeignKey):
            related_original = getattr(original_obj, field_name)
            related_translated = handle_related_object(related_original, src_locale, tgt_locale, field_segments)
            setattr(translated_obj, field_name, related_translated)

        elif isinstance(field, (models.ManyToOneRel)):
            original_manager = getattr(original_obj, field_name)
            translated_manager = getattr(translated_obj, field_name)

            segments_by_child = defaultdict(list)

            for segment in field_segments:
                child_translation_key, segment = segment.unwrap()
                segments_by_child[child_translation_key].append(segment)

            for child_translation_key, child_segments in segments_by_child.items():
                original_child_object = original_manager.filter(translation_key=child_translation_key).first()
                translated_child_object = translated_manager.filter(translation_key=child_translation_key).first()

                if not translated_child_object:
                    # TODO: Here, we expect that the inline child to already exist as Wagtail copies it
                    # when creating the translated page. When we add editing, we will need to support
                    # adding new inline objects manually.
                    continue

                ingest_segments(original_child_object, translated_child_object, src_locale, tgt_locale, child_segments)
                translated_child_object.save()



from django.conf.urls import include, url

from wagtail.core import hooks

from wagtail_localize.admin.workflow.action_modules import BaseActionModule

from .views import download, upload


class DownloadUploadPOFileActionModule(BaseActionModule):
    template_name = 'wagtail_localize_pofile/action_module.html'


@hooks.register('wagtail_localize_workflow_register_action_modules')
def wagtail_localize_workflow_register_action_modules():
    return [
        DownloadUploadPOFileActionModule,
    ]


@hooks.register('register_admin_urls')
def register_admin_urls():
    urls = [
        url('^download/(\d+)/$', download, name='download'),
        url('^upload/(\d+)/$', upload, name='upload'),
    ]

    return [
        url('^localize/engine/pofile/', include((urls, 'wagtail_localize_pofile'), namespace='wagtail_localize_pofile')),
    ]



from django.core.cache import cache
from django.db import models
from django.db.models import Case, Q, When
from django.utils.translation import ugettext_lazy as _

from wagtail_localize.models import Language, Region


MATCH_HOSTNAME_PORT = 0
MATCH_HOSTNAME_DEFAULT = 1
MATCH_DEFAULT = 2
MATCH_HOSTNAME = 3


class SiteManager(models.Manager):
    def get_by_natural_key(self, hostname, port):
        return self.get(hostname=hostname, port=port)


class Site(models.Model):
    hostname = models.CharField(verbose_name=_('hostname'), max_length=255, db_index=True)
    port = models.IntegerField(
        verbose_name=_('port'),
        default=80,
        help_text=_(
            "Set this to something other than 80 if you need a specific port number to appear in URLs"
            " (e.g. development on port 8000). Does not affect request handling (so port forwarding still works)."
        )
    )
    site_name = models.CharField(
        verbose_name=_('site name'),
        max_length=255,
        null=True,
        blank=True,
        help_text=_("Human-readable name for the site.")
    )
    is_default_site = models.BooleanField(
        verbose_name=_('is default site'),
        default=False,
        help_text=_(
            "If true, this site will handle requests for all other hostnames that do not have a site entry of their own"
        )
    )

    objects = SiteManager()

    class Meta:
        unique_together = ('hostname', 'port')
        verbose_name = _('site')
        verbose_name_plural = _('sites')

    def natural_key(self):
        return (self.hostname, self.port)

    def __str__(self):
        if self.site_name:
            return(
                self.site_name
                + (" [default]" if self.is_default_site else "")
            )
        else:
            return(
                self.hostname
                + ("" if self.port == 80 else (":%d" % self.port))
                + (" [default]" if self.is_default_site else "")
            )

    @classmethod
    def find_for_request(cls, request):
        """
        Find the site object responsible for responding to this HTTP
        request object. Try:

        * unique hostname first
        * then hostname and port
        * if there is no matching hostname at all, or no matching
          hostname:port combination, fall back to the unique default site,
          or raise an exception

        NB this means that high-numbered ports on an extant hostname may
        still be routed to a different hostname which is set as the default
        """

        hostname = request.get_host().split(':')[0]
        port = request.get_port()

        sites = list(cls.objects.annotate(match=Case(
            # annotate the results by best choice descending

            # put exact hostname+port match first
            When(hostname=hostname, port=port, then=MATCH_HOSTNAME_PORT),

            # then put hostname+default (better than just hostname or just default)
            When(hostname=hostname, is_default_site=True, then=MATCH_HOSTNAME_DEFAULT),

            # then match default with different hostname. there is only ever
            # one default, so order it above (possibly multiple) hostname
            # matches so we can use sites[0] below to access it
            When(is_default_site=True, then=MATCH_DEFAULT),

            # because of the filter below, if it's not default then its a hostname match
            default=MATCH_HOSTNAME,

            output_field=models.IntegerField(),
        )).filter(Q(hostname=hostname) | Q(is_default_site=True)).order_by(
            'match'
        ))

        if sites:
            # if theres a unique match or hostname (with port or default) match
            if len(sites) == 1 or sites[0].match in (MATCH_HOSTNAME_PORT, MATCH_HOSTNAME_DEFAULT):
                return sites[0]

            # if there is a default match with a different hostname, see if
            # there are many hostname matches. if only 1 then use that instead
            # otherwise we use the default
            if sites[0].match == MATCH_DEFAULT:
                return sites[len(sites) == 2]

        raise cls.DoesNotExist()

    @property
    def root_page(self):
        try:
            language = Language.get_active()
            return self.languages.get(language=language).root_page
        except (Language.DoesNotExist, SiteLanguage.DoesNotExist):
            language = Language.default()
            return self.languages.get(language=language).root_page

    @property
    def root_url(self):
        if self.port == 80:
            return 'http://%s' % self.hostname
        elif self.port == 443:
            return 'https://%s' % self.hostname
        else:
            return 'http://%s:%d' % (self.hostname, self.port)

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        # Only one site can have the is_default_site flag set
        try:
            default = Site.objects.get(is_default_site=True)
        except Site.DoesNotExist:
            pass
        except Site.MultipleObjectsReturned:
            raise
        else:
            if self.is_default_site and self.pk != default.pk:
                raise ValidationError(
                    {'is_default_site': [
                        _(
                            "%(hostname)s is already configured as the default site."
                            " You must unset that before you can save this site as default."
                        )
                        % {'hostname': default.hostname}
                    ]}
                )

    @staticmethod
    def get_site_root_paths():
        """
        Return a list of (id, root_path, root_url) tuples, most specific path
        first - used to translate url_paths into actual URLs with hostnames
        """
        result = cache.get('wagtail_site_root_paths')

        if result is None:
            result = []
            for site in Site.objects.order_by( '-is_default_site', 'hostname'):
                result.extend(
                    (site.id, site_language.root_page.url_path, site.root_url, site_language.language.code)
                    for site_language in site.languages.filter(is_active=True)
                )

            cache.set('wagtail_site_root_paths', result, 3600)

        return result


def default_region_id():
    return Region.objects.default_id()


class SiteLanguage(models.Model):
    site = models.ForeignKey(Site, on_delete=models.CASCADE, related_name='languages')
    region = models.ForeignKey(Region, verbose_name=_('region'), on_delete=models.CASCADE, related_name='+', default=default_region_id)
    language = models.ForeignKey(Language, verbose_name=_('language'), on_delete=models.CASCADE, related_name='+')
    root_page = models.ForeignKey('wagtailcore.Page', verbose_name=_('root page'), on_delete=models.CASCADE, related_name='+')
    is_active = models.BooleanField(_('is active'), default=True)



from django.contrib.auth.models import Permission
from django.conf.urls import url, include
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from wagtail.admin import widgets as wagtailadmin_widgets
from wagtail.admin.menu import MenuItem
from wagtail.core import hooks
from wagtail.core.permission_policies import ModelPermissionPolicy

from wagtail_localize.models import TranslatablePageMixin

from .models import TranslationRequest
from .views.create_translation_request import create_translation_request
from .views.management import TranslationRequestViewSet


@hooks.register('register_page_listing_more_buttons')
def page_listing_more_buttons(page, page_perms, is_parent=False):
    if isinstance(page, TranslatablePageMixin):
        yield wagtailadmin_widgets.Button(
            'Create translation request',
            reverse('wagtail_localize_workflow:create_translation_request', args=[page.id]),
            priority=60
        )


@hooks.register('register_admin_urls')
def register_admin_urls():
    urls = [
        url('^create_translation_request/(\d+)/$', create_translation_request, name='create_translation_request'),
    ]

    return [
        url('^localize/workflow/', include((urls, 'wagtail_localize_workflow'), namespace='wagtail_localize_workflow')),
    ]


@hooks.register('register_admin_viewset')
def register_viewset():
    return TranslationRequestViewSet('wagtail_localize_workflow_management', url_prefix='localize/workflow/translations')


class TranslationsMenuItem(MenuItem):
    def is_shown(self, request):
        return True


@hooks.register('register_admin_menu_item')
def register_menu_item():
    return TranslationsMenuItem(_('Translations'), reverse('wagtail_localize_workflow_management:list'),
                         classnames='icon icon-site', order=500)



from django.apps import AppConfig


class WagtailLocalizeTestsAppConfig(AppConfig):
    label = 'wagtail_localize_tests'
    name = 'wagtail_localize.tests'
    verbose_name = "Wagtail Localize Tests"



from bs4 import BeautifulSoup, NavigableString


INLINE_TAGS = ['a', 'abbr', 'acronym', 'b', 'code', 'em', 'i', 'strong']


def lstrip_keep(text):
    """
    Like lstrip, but also returns the whitespace that was stripped off
    """
    text_length = len(text)
    new_text = text.lstrip()
    prefix = text[0:(text_length-len(new_text))]
    return new_text, prefix


def rstrip_keep(text):
    """
    Like rstrip, but also returns the whitespace that was stripped off
    """
    text_length = len(text)
    new_text = text.rstrip()
    if text_length != len(new_text):
        suffix = text[-(text_length-len(new_text)):]
    else:
        suffix = ''
    return new_text, suffix


def extract_html_segments(html):
    """
    This function extracts translatable segments from an HTML fragment.

    Inline elements and visible text are extracted together.

    For example:

        <h1>Foo</h1>
        <p>
            Bar
            <ul>
                <li><b>Baz</b></li>
            </ul>
        </p>

    Will produce the following two outputs (as a 2-tuple)

        <h1><text position="0"></h1>
        <p>
            <text position="1">
            <ul>
                <li><text position="2"></li>
            </ul>
        </p>

        [
            "Foo",
            "Bar",
            "<b>Baz</b>",
        ]
    """
    soup = BeautifulSoup(html, 'html.parser')

    def wrap(elements):
        """
        Wraps the given elements with a <text> tag

        The elements must be contiguous siblings or this might screw up the tree.
        """
        value = ''.join(element.output_ready() if isinstance(element, NavigableString) else str(element) for element in elements)

        if value and not value.isspace():
            # Create <text> tag
            elements[0].insert_before(soup.new_tag('text', value=value))

            # Remove elements
            for element in elements:
                element.replaceWith('')

    def walk(element):
        """
        Walks the tree in depth first search post-order.

        When it encounters an element that could be extracted, it wraps it with
        a <text> tag. These are extracted in the next stage (because we want to
        preserve order of occurance).

        For example:

        <p>
            Foo
            <ul>
              <li>Bar</li>
            </ul>
            Baz
        </p>

        Is transformed to:

        <p>
            <text>Foo</text>
            <ul>
              <li><text><b>Bar</b></text></li>
            </ul>
            <text>Baz</text>
        </p>
        """
        if isinstance(element, NavigableString):
            return False

        has_block = False
        buffer = []

        for child in element.children:
            is_block = walk(child)

            if is_block:
                has_block = True

                if buffer:
                    wrap(buffer)
                    buffer = []

            else:
                buffer.append(child)

        if buffer and has_block:
            wrap(buffer)
            buffer = []

        if element.name not in INLINE_TAGS:
            if buffer:
               wrap(buffer)

            return True

        return False

    walk(soup)

    # Now extract segments from the <text> tags
    segments = []
    for element in soup.descendants:
        if element.name == 'text':
            text = element.attrs.pop('value')

            # Strip leading and trailing whitespace. We keep the values and reinsert them
            # into the template
            # This is probably not necessary, but just to be on the safe side
            text, prefix = lstrip_keep(text)
            text, suffix = rstrip_keep(text)

            element.attrs['position'] = len(segments)
            segments.append(text.strip())

            if prefix:
                element.insert_before(prefix)

            if suffix:
                element.insert_after(suffix)

    return str(soup), segments


def restore_html_segments(template, texts):
    soup = BeautifulSoup(template, 'html.parser')

    for text_element in soup.findAll('text'):
        value = texts[int(text_element.get('position'))]
        text_element.replaceWith(value.strip())

    return str(soup)


def extract_html_elements(html):
    """
    Extracts HTML elements from a fragment. Returns the plain text representation
    of the HTML document and an array of elements including their span, type and attributes.

    For example:

    text, elements = extract_html_elements("This is a paragraph. <b>This is some bold <i>and now italic</i></b> text")

    text == "This is a paragraph. This is some bold and now italic text"
    elements == [(39, 53, 'i', {}), (21, 53, 'b', {})]
    """
    soup = BeautifulSoup(html, 'html.parser')

    texts = []
    cursor = {'current': 0}
    elements = []

    def walk(soup):
        for element in soup.children:
            if isinstance(element, NavigableString):
               texts.append(element)
               cursor['current'] += len(element)

            else:
                start = cursor['current']
                walk(element)
                end = cursor['current']

                elements.append((start, end, element.name, element.attrs.copy()))

    walk(soup)

    return ''.join(texts), elements


def restore_html_elements(text, elements):
    """
    Inserts elements into a plain text string returning a HTML document.
    """
    soup = BeautifulSoup('', 'html.parser')
    stack = []
    cursor = 0
    current_element = soup

    # Sort elements by start position
    elements.sort(key=lambda element: element[0])

    for i, element in enumerate(elements):
        if cursor < element[0]:
            # Output text and advance cursor
            current_element.append(text[cursor:element[0]])
            cursor = element[0]

        stack.append((element[1], current_element))
        new_element = soup.new_tag(element[2], **element[3])
        current_element.append(new_element)
        current_element = new_element

        # Close existing elements before going to the next element
        while stack:
            if i < len(elements) - 1:
                if stack[len(stack) - 1][0] > elements[i+1][0]:
                    # New element created before this one closes.
                    # Go to next element
                    break

            element_end, previous_element = stack.pop()

            if cursor < element_end:
                # Output text and advance cursor
                current_element.append(text[cursor:element_end])
                cursor = element_end

            current_element = previous_element

    if cursor < len(text):
        current_element.append(text[cursor:])

    return str(soup)



# Generated by Django 2.1.9 on 2019-06-21 09:16

from django.db import migrations, models
import django.db.models.deletion
import wagtail_localize.experimental.sites.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtail_localize', '0001_initial'),
        ('wagtailcore', '0040_page_draft_title'),
    ]

    operations = [
        migrations.CreateModel(
            name='Site',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hostname', models.CharField(db_index=True, max_length=255, verbose_name='hostname')),
                ('port', models.IntegerField(default=80, help_text='Set this to something other than 80 if you need a specific port number to appear in URLs (e.g. development on port 8000). Does not affect request handling (so port forwarding still works).', verbose_name='port')),
                ('site_name', models.CharField(blank=True, help_text='Human-readable name for the site.', max_length=255, null=True, verbose_name='site name')),
                ('is_default_site', models.BooleanField(default=False, help_text='If true, this site will handle requests for all other hostnames that do not have a site entry of their own', verbose_name='is default site')),
            ],
            options={
                'verbose_name': 'site',
                'verbose_name_plural': 'sites',
            },
        ),
        migrations.CreateModel(
            name='SiteLanguage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, verbose_name='is active')),
                ('language', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtail_localize.Language', verbose_name='language')),
                ('region', models.ForeignKey(default=wagtail_localize.experimental.sites.models.default_region_id, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtail_localize.Region', verbose_name='region')),
                ('root_page', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtailcore.Page', verbose_name='root page')),
                ('site', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='languages', to='wagtail_localize_sites.Site')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='site',
            unique_together={('hostname', 'port')},
        ),
    ]



from django.apps import AppConfig


class WagtailLocalizeTranslationMemoryAppConfig(AppConfig):
    label = 'wagtail_localize_translation_memory'
    name = 'wagtail_localize.translation_memory'
    verbose_name = "Wagtail Localize Translation Memory"




from django.core.management.base import BaseCommand, CommandError

from wagtail_localize.bootstrap import bootstrap_translatable_model
from wagtail_localize.models import BootstrapTranslatableMixin, get_translatable_models


class Command(BaseCommand):
    help = 'Populates require fields of new translatable models'

    def handle(self, **options):
        for model in get_translatable_models():
            if issubclass(model, BootstrapTranslatableMixin):
                print("Bootstrapping: {}.{}".format(model._meta.app_label, model.__name__))

                bootstrap_translatable_model(model)



from django.conf.urls import url
from django.templatetags.static import static
from django.utils.html import format_html_join

from wagtail.core import hooks

from .views import translations_list


@hooks.register('insert_editor_js')
def insert_editor_js():
    js_files = [
        'wagtail_localize_language_switch/js/language-switch.js',
    ]
    return format_html_join('\n', '<script src="{0}"></script>',
        ((static(filename),) for filename in js_files)
    )


@hooks.register('register_admin_urls')
def register_admin_urls():
    return [
        url(r'^localize/language_switch/(\d+)/$', translations_list, name='translations_list'),
    ]



import uuid
import unittest

from django.test import TestCase

from wagtail.core.blocks import StreamValue
from wagtail.core.models import Page

from wagtail_localize.tests.models import TestPage, TestSnippet, TestChildObject
from wagtail_localize.segments import SegmentValue, TemplateValue
from wagtail_localize.segments.extract import extract_segments


def make_test_page(**kwargs):
    root_page = Page.objects.get(id=1)
    kwargs.setdefault('title', "Test page")
    return root_page.add_child(instance=TestPage(**kwargs))


RICH_TEXT_TEST_INPUT = '<h1>This is a heading</h1><p>This is a paragraph. &lt;foo&gt; <b>Bold text</b></p><ul><li><a href="http://example.com">This is a link</a></li></ul>'

RICH_TEXT_TEST_OUTPUT = [
    TemplateValue(
        '',
        'html',
        '<h1><text position="0"></text></h1><p><text position="1"></text></p><ul><li><text position="2"></text></li></ul>',
        3
    ),
    SegmentValue('0', "This is a heading", html_elements=[]),
    SegmentValue('1', "This is a paragraph. <foo> Bold text", html_elements=[
        SegmentValue.HTMLElement(27, 36, 'b1', ('b', {})),
    ]),
    SegmentValue('2', "This is a link", html_elements=[
        SegmentValue.HTMLElement(0, 14, 'a1', ('a', {'href': 'http://example.com'})),
    ]),
]


class TestSegmentExtraction(TestCase):
    def test_charfield(self):
        page = make_test_page(test_charfield="Test content")
        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_charfield', "Test content"),
        ])

    def test_textfield(self):
        page = make_test_page(test_textfield="Test content")
        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_textfield', "Test content"),
        ])

    def test_emailfield(self):
        page = make_test_page(test_emailfield="test@example.com")
        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_emailfield', "test@example.com"),
        ])

    def test_slugfield(self):
        page = make_test_page(test_slugfield="test-content")
        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_slugfield', "test-content"),
        ])

    def test_urlfield(self):
        page = make_test_page(test_urlfield="http://test-content.com/foo")
        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_urlfield', "http://test-content.com/foo"),
        ])

    def test_richtextfield(self):
        page = make_test_page(test_richtextfield=RICH_TEXT_TEST_INPUT)
        segments = extract_segments(page)

        self.assertEqual(segments, [segment.wrap('test_richtextfield') for segment in RICH_TEXT_TEST_OUTPUT])

    def test_snippet(self):
        test_snippet = TestSnippet.objects.create(field="Test content")
        page = make_test_page(test_snippet=test_snippet)
        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_snippet.field', "Test content"),
        ])

    def test_childobjects(self):
        page = make_test_page()
        page.test_childobjects.add(TestChildObject(field="Test content"))
        page.save()

        child_translation_key = TestChildObject.objects.get().translation_key

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_childobjects.{child_translation_key}.field', "Test content"),
        ])

    def test_customfield(self):
        page = make_test_page(test_customfield="Test content")

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue('test_customfield.foo', "Test content and some extra"),
        ])


def make_test_page_with_streamfield_block(block_id, block_type, block_value, **kwargs):
    stream_data = [
        {
            'id': block_id,
            'type': block_type,
            'value': block_value
        }
    ]

    return make_test_page(
        test_streamfield=StreamValue(TestPage.test_streamfield.field.stream_block, stream_data, is_lazy=True),
        **kwargs
    )


class TestSegmentExtractionWithStreamField(TestCase):
    def test_charblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_charblock', "Test content")

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}', "Test content"),
        ])

    def test_textblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_textblock', "Test content")

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}', "Test content"),
        ])

    @unittest.expectedFailure  # Not supported
    def test_emailblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_emailblock', "test@example.com")

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}', "test@example.com"),
        ])

    @unittest.expectedFailure  # Not supported
    def test_urlblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_urlblock', "http://test-content.com/foo")

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}', "http://test-content.com/foo"),
        ])

    def test_richtextblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_richtextblock', RICH_TEXT_TEST_INPUT)

        segments = extract_segments(page)

        self.assertEqual(segments, [segment.wrap(f'test_streamfield.{block_id}') for segment in RICH_TEXT_TEST_OUTPUT])

    @unittest.expectedFailure  # Not supported
    def test_rawhtmlblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_rawhtmlblock', RICH_TEXT_TEST_INPUT)

        segments = extract_segments(page)

        self.assertEqual(segments, [segment.wrap(f'test_streamfield.{block_id}') for segment in RICH_TEXT_TEST_OUTPUT])

    def test_blockquoteblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_blockquoteblock', "Test content")

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}', "Test content"),
        ])

    def test_structblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_structblock', {
            'field_a': "Test content",
            'field_b': "Some more test content",
        })

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}.field_a', "Test content"),
            SegmentValue(f'test_streamfield.{block_id}.field_b', "Some more test content"),
        ])

    @unittest.expectedFailure  # Not supported (probably won't ever be due to lack of path stability)
    def test_listblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_listblock', [
            "Test content",
            "Some more test content",
        ])

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}.0', "Test content"),
            SegmentValue(f'test_streamfield.{block_id}.1', "Some more test content"),
        ])

    @unittest.expectedFailure  # Not supported
    def test_nestedstreamblock(self):
        block_id = uuid.uuid4()
        nested_block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_nestedstreamblock', [
            {
                'id': str(nested_block_id),
                'type': 'block_a',
                'value': "Test content",
            }
        ])

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}.{nested_block_id}', "Test content"),
        ])

    def test_customstructblock(self):
        block_id = uuid.uuid4()
        page = make_test_page_with_streamfield_block(str(block_id), 'test_customstructblock', {
            'field_a': "Test content",
            'field_b': "Some more test content",
        })

        segments = extract_segments(page)

        self.assertEqual(segments, [
            SegmentValue(f'test_streamfield.{block_id}.foo', "Test content Some more test content"),
        ])



# Generated by Django 2.1.9 on 2019-06-21 09:16

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('wagtail_localize', '0001_initial'),
        ('wagtailcore', '0040_page_draft_title'),
    ]

    operations = [
        migrations.CreateModel(
            name='TranslationRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField()),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='translation_requests_created', to=settings.AUTH_USER_MODEL)),
                ('source_locale', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtail_localize.Locale')),
                ('target_locale', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtail_localize.Locale')),
                ('target_root', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtailcore.Page')),
            ],
        ),
        migrations.CreateModel(
            name='TranslationRequestPage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_completed', models.BooleanField(default=False)),
                ('completed_revision', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.PageRevision')),
                ('parent', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='child_pages', to='wagtail_localize_workflow.TranslationRequestPage')),
                ('request', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pages', to='wagtail_localize_workflow.TranslationRequest')),
                ('source_revision', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='wagtailcore.PageRevision')),
            ],
        ),
    ]



# Generated by Django 2.1.10 on 2019-08-13 12:30

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtail_localize', '0002_initial_data'),
        ('wagtailcore', '0041_group_collection_permissions_verbose_name_plural'),
    ]

    operations = [
        migrations.CreateModel(
            name='Segment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(unique=True)),
                ('text', models.TextField()),
                ('language', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtail_localize.Language')),
            ],
        ),
        migrations.CreateModel(
            name='SegmentPageLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path', models.TextField()),
                ('page_revision', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.PageRevision')),
                ('segment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='page_locations', to='wagtail_localize_translation_memory.Segment')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SegmentTranslation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField()),
                ('language', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtail_localize.Language')),
                ('translation_of', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='wagtail_localize_translation_memory.Segment')),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(unique=True)),
                ('template', models.TextField()),
                ('template_format', models.CharField(max_length=100)),
                ('segment_count', models.PositiveIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='TemplatePageLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path', models.TextField()),
                ('page_revision', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.PageRevision')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='page_locations', to='wagtail_localize_translation_memory.Template')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterUniqueTogether(
            name='segmenttranslation',
            unique_together={('language', 'translation_of')},
        ),
    ]



from django.db import transaction
from django.utils.translation import ugettext_lazy

from wagtail.admin.views import generic
from wagtail.admin.viewsets.model import ModelViewSet
from wagtail.core.permission_policies import ModelPermissionPolicy

from wagtail_localize.models import Region
from .components import get_region_components, get_region_component_edit_handler
from .forms import RegionForm


class RegionComponentManager:
    def __init__(self, components):
        self.components = components

    @classmethod
    def from_request(cls, request, instance=None):
        components = []

        for component_model in get_region_components():
            component_instance = component_model.objects.filter(region=instance).first()
            edit_handler = get_region_component_edit_handler(component_model).bind_to(model=component_model, instance=component_instance, request=request)
            form_class = edit_handler.get_form_class()
            prefix = f'component_{component_model._meta.app_label}_{component_model.__name__}'

            if request.method == 'POST':
                form = form_class(request.POST, request.FILES, instance=component_instance, prefix=prefix)
            else:
                form = form_class(instance=component_instance, prefix=prefix)

            components.append((component_model, component_instance, form))

        return cls(components)

    def is_valid(self):
        return all(component_form.is_valid() for component_model, component_instance, component_form in self.components)

    def save(self, region):
        for component_model, component_instance, component_form in self.components:
            component_instance = component_form.save(commit=False)
            component_instance.region = region
            component_instance.save()

    def __iter__(self):
        return iter(self.components)


class IndexView(generic.IndexView):
    template_name = 'wagtail_localize_regions_admin/index.html'
    page_title = ugettext_lazy("Regions")
    add_item_label = ugettext_lazy("Add a region")
    context_object_name = 'regions'


class CreateView(generic.CreateView):
    page_title = ugettext_lazy("Add region")
    success_message = ugettext_lazy("Region '{0}' created.")
    template_name = 'wagtail_localize_regions_admin/create.html'

    def get_components(self):
        return RegionComponentManager.from_request(self.request)

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        self.components = RegionComponentManager.from_request(self.request)

        if form.is_valid() and self.get_components().is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    @transaction.atomic
    def form_valid(self, form):
        response = super().form_valid(form)
        self.get_components().save(self.object)
        return response

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        context['components'] = self.get_components()
        return context


class EditView(generic.EditView):
    success_message = ugettext_lazy("Region '{0}' updated.")
    error_message = ugettext_lazy("The site could not be saved due to errors.")
    delete_item_label = ugettext_lazy("Delete region")
    context_object_name = 'region'
    template_name = 'wagtail_localize_regions_admin/edit.html'

    def get_components(self):
        return RegionComponentManager.from_request(self.request, instance=self.object)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()

        if form.is_valid() and self.get_components().is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    @transaction.atomic
    def form_valid(self, form):
        response = super().form_valid(form)
        self.get_components().save(self.object)
        return response

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        context['components'] = self.get_components()
        return context


class DeleteView(generic.DeleteView):
    success_message = ugettext_lazy("Region '{0}' deleted.")
    page_title = ugettext_lazy("Delete region")
    confirmation_message = ugettext_lazy("Are you sure you want to delete this region?")


class RegionViewSet(ModelViewSet):
    icon = 'site'
    model = Region
    permission_policy = ModelPermissionPolicy(Region)

    index_view_class = IndexView
    add_view_class = CreateView
    edit_view_class = EditView
    delete_view_class = DeleteView

    def get_form_class(self, for_update=False):
        return RegionForm



from collections import defaultdict

from django.contrib import messages
from django.db import transaction
from django.views.decorators.http import require_POST
from django.shortcuts import get_object_or_404, redirect
from django.utils.text import slugify

from wagtail_localize.admin.workflow.models import TranslationRequest
from wagtail_localize.segments import SegmentValue, TemplateValue
from wagtail_localize.segments.extract import extract_segments
from wagtail_localize.segments.ingest import ingest_segments

# TODO: Switch to official Google API client
from googletrans import Translator


def language_code(code):
    if code in ['zh-hans', 'zh-cn']:
        return 'zh-cn'

    if code in ['zh-hant', 'zh-tw']:
        return 'zh-tw'

    return code.split('-')[0]


@require_POST
def translate(request, translation_request_id):
    translation_request = get_object_or_404(TranslationRequest, id=translation_request_id)
    translator = Translator()

    for page in translation_request.pages.filter(is_completed=False):
        instance = page.source_revision.as_page_object()

        segments = extract_segments(instance)

        text_segments = [segment for segment in segments if isinstance(segment, SegmentValue)]
        template_segments = [segment for segment in segments if isinstance(segment, TemplateValue)]

        # Group segments by source text so we only submit them once
        text_segments_grouped = defaultdict(list)
        for segment in text_segments:
            text_segments_grouped[segment.text].append(segment.path)

        translations = translator.translate(
            list(text_segments_grouped.keys()),
            src=language_code(translation_request.source_locale.language.code),
            dest=language_code(translation_request.target_locale.language.code),
        )

        translated_segments = template_segments.copy()
        for translation in translations:
            translated_segments.extend([
                SegmentValue(path, translation.text)
                for path in text_segments_grouped[translation.origin]
            ])

        with transaction.atomic():
            try:
                translation = instance.get_translation(translation_request.target_locale)
            except instance.__class__.DoesNotExist:
                translation = instance.copy_for_translation(translation_request.target_locale)

            ingest_segments(instance, translation, translation_request.source_locale, translation_request.target_locale, translated_segments)
            translation.slug = slugify(translation.slug)
            revision = translation.save_revision()

            # Update translation request
            page.is_completed = True
            page.completed_revision = revision
            page.save(update_fields=['is_completed', 'completed_revision'])

    # TODO: Plural
    messages.success(request, "%d pages successfully translated with Google Translate" % translation_request.pages.count())

    return redirect('wagtail_localize_workflow_management:detail', translation_request_id)



from django.db import models, transaction
from django.db.models import OuterRef, Prefetch, Subquery
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.text import slugify

from wagtail.core.models import Page, PageRevision

from wagtail_localize.models import Language
from wagtail_localize.segments import SegmentValue, TemplateValue
from wagtail_localize.segments.ingest import ingest_segments

from .models import Segment, SegmentTranslation, SegmentPageLocation, TemplatePageLocation
from .utils import get_translation_progress


@transaction.atomic
def handle_completed_revision(revision_id, src_lang, tgt_lang):
    original_page = Page.objects.get(revisions__id=revision_id).specific

    # TODO: Not currently handling edited page translations
    if original_page.has_translation(tgt_lang):
        return

    original_page_at_revision = original_page.revisions.get(id=revision_id).as_page_object()
    translated_page = original_page_at_revision.copy_for_translation(tgt_lang)

    # If the page doesn't yet have a translated parent, don't import it yet
    # This function will be called again when the translated parent is created
    if not translated_page:
        return

    # Fetch all translated segments
    segment_page_locations = (
        SegmentPageLocation.objects
        .filter(page_revision_id=revision_id)
        .annotate_translation(tgt_lang)
    )

    template_page_locations = (
        TemplatePageLocation.objects
        .filter(page_revision_id=revision_id)
        .select_related('template')
    )

    segments = []

    for page_location in segment_page_locations:
        segment = SegmentValue(page_location.path, page_location.translation)
        segments.append(segment)

    for page_location in template_page_locations:
        template = page_location.template
        segment = TemplateValue(page_location.path, template.template_format, template.template, template.segment_count)
        segments.append(segment)

    ingest_segments(original_page_at_revision, translated_page, src_lang, tgt_lang, segments)

    translated_page.slug = slugify(translated_page.slug)

    translated_page.save()

    # Create initial revision
    translated_page.save_revision(changed=False)

    # As we created a new page, check to see if there are any child pages that
    # have complete translations as well. This can happen if the child pages
    # were translated before the parent.

    # Check all page revisions of children of the original page that have
    # been submitted for localisation
    child_page_revision_ids = PageRevision.objects.filter(page__in=original_page.get_children()).values_list('id', flat=True)
    submitted_page_revision_ids = SegmentPageLocation.objects.values_list('page_revision_id', flat=True)
    page_revision_ids_to_check = child_page_revision_ids.intersection(submitted_page_revision_ids)

    for page_revision_id in page_revision_ids_to_check:
        total_segments, translated_segments = get_translation_progress(page_revision_id, tgt_lang)

        if total_segments == translated_segments:
            # Page revision is now complete
            handle_completed_revision(page_revision_id, src_lang, tgt_lang)


@receiver(post_save, sender=SegmentTranslation)
def on_new_segment_translation(sender, instance, created, **kwargs):
    if created:
        segment_page_locations = SegmentPageLocation.objects.filter(segment_id=instance.translation_of_id)

        # Check if any translated pages can now be created
        for page_revision_id in segment_page_locations.values_list('page_revision_id', flat=True):
            total_segments, translated_segments = get_translation_progress(page_revision_id, instance.locale.language)

            if total_segments == translated_segments:
                # FIXME
                src_lang = Language.default()

                # Page revision is now complete
                handle_completed_revision(page_revision_id, src_lang, instance.locale.language)


@transaction.atomic
def ingest_translations(src_lang, tgt_lang, translations):
    for source, translation in translations:
        segment = Segment.from_text(src_lang, source)
        SegmentTranslation.from_text(segment, tgt_lang, translation)



from django.db import models

from modelcluster.fields import ParentalKey
from wagtail.core import blocks
from wagtail.core.fields import RichTextField, StreamField
from wagtail.core.models import Page, Orderable

from wagtail_localize.models import TranslatableMixin, TranslatablePageMixin
from wagtail_localize.segments import SegmentValue


class TestSnippet(TranslatableMixin, models.Model):
    field = models.TextField()

    translatable_fields = ['field']


class TestStructBlock(blocks.StructBlock):
    field_a = blocks.TextBlock()
    field_b = blocks.TextBlock()


class TestNestedStreamBlock(blocks.StreamBlock):
    block_a = blocks.TextBlock()
    block_b = blocks.TextBlock()


class CustomStructBlock(blocks.StructBlock):
    field_a = blocks.TextBlock()
    field_b = blocks.TextBlock()

    def get_translatable_segments(self, value):
        return [
            SegmentValue('foo', f"{value['field_a']} {value['field_b']}"),
        ]


class TestStreamBlock(blocks.StreamBlock):
    test_charblock = blocks.CharBlock(max_length=255)
    test_textblock = blocks.TextBlock()
    test_emailblock = blocks.EmailBlock()
    test_urlblock = blocks.URLBlock()
    test_richtextblock = blocks.RichTextBlock()
    test_rawhtmlblock = blocks.RawHTMLBlock()
    test_blockquoteblock = blocks.BlockQuoteBlock()
    test_structblock = TestStructBlock()
    test_listblock = blocks.ListBlock(blocks.TextBlock())
    test_nestedstreamblock = TestNestedStreamBlock()
    test_customstructblock = CustomStructBlock()


class TestCustomField(models.TextField):
    def get_translatable_segments(self, value):
        if not value:
            # Don't disrupt other tests
            return []

        return [
            SegmentValue('foo', f"{value} and some extra"),
        ]


class TestPage(TranslatablePageMixin, Page):
    test_charfield = models.CharField(max_length=255, blank=True)
    test_textfield = models.TextField(blank=True)
    test_emailfield = models.EmailField(blank=True)
    test_slugfield = models.SlugField(blank=True)
    test_urlfield = models.URLField(blank=True)

    test_richtextfield = RichTextField(blank=True)
    test_streamfield = StreamField(TestStreamBlock, blank=True)

    test_snippet = models.ForeignKey(TestSnippet, null=True, blank=True, on_delete=models.SET_NULL)

    test_customfield = TestCustomField(blank=True)

    translatable_fields = [
        'test_charfield',
        'test_textfield',
        'test_emailfield',
        'test_slugfield',
        'test_urlfield',
        'test_richtextfield',
        'test_streamfield',
        'test_snippet',
        'test_childobjects',
        'test_customfield',
    ]


class TestChildObject(TranslatableMixin, Orderable):
    page = ParentalKey(TestPage, related_name='test_childobjects')
    field = models.TextField()

    translatable_fields = ['field']



import uuid

from django.db import models, transaction
from django.db.models import Subquery, OuterRef


def pk(obj):
    if isinstance(obj, models.Model):
        return obj.pk
    else:
        return obj


class SegmentQuerySet(models.QuerySet):
    def annotate_translation(self, language):
        """
        Adds a 'translation' field to the segments containing the
        text content of the segment translated into the specified
        language.
        """
        return self.annotate(
            translation=Subquery(
                SegmentTranslation.objects.filter(
                    translation_of_id=OuterRef('pk'),
                    language_id=pk(language),
                ).values('text')
            )
        )


class Segment(models.Model):
    UUID_NAMESPACE = uuid.UUID('59ed7d1c-7eb5-45fa-9c8b-7a7057ed56d7')

    language = models.ForeignKey('wagtail_localize.Language', on_delete=models.CASCADE)
    uuid = models.UUIDField(unique=True)
    text = models.TextField()

    objects = SegmentQuerySet.as_manager()

    @classmethod
    def from_text(cls, language, text):
        segment, created = cls.objects.get_or_create(
            language_id=pk(language),
            uuid=uuid.uuid5(cls.UUID_NAMESPACE, text),
            defaults={
                'text': text,
            }
        )

        return segment


class SegmentTranslation(models.Model):
    translation_of = models.ForeignKey(Segment, on_delete=models.CASCADE, related_name='translations')
    language = models.ForeignKey('wagtail_localize.Language', on_delete=models.CASCADE)
    text = models.TextField()

    class Meta:
        unique_together = [
            ('language', 'translation_of'),
        ]

    @classmethod
    def from_text(cls, translation_of, language, text):
        segment, created = cls.objects.get_or_create(
            translation_of=translation_of,
            language_id=pk(language),
            defaults={
                'text': text,
            }
        )

        return segment


class Template(models.Model):
    BASE_UUID_NAMESPACE = uuid.UUID('4599eabc-3f8e-41a9-be61-95417d26a8cd')

    uuid = models.UUIDField(unique=True)
    template = models.TextField()
    template_format = models.CharField(max_length=100)
    segment_count = models.PositiveIntegerField()

    @classmethod
    def from_template_value(cls, template_value):
        uuid_namespace = uuid.uuid5(cls.BASE_UUID_NAMESPACE, template_value.format)

        template, created = cls.objects.get_or_create(
            uuid=uuid.uuid5(uuid_namespace, template_value.template),
            defaults={
                'template': template_value.template,
                'template_format': template_value.format,
                'segment_count': template_value.segment_count,
            }
        )

        return template


class BasePageLocation(models.Model):
    page_revision = models.ForeignKey('wagtailcore.PageRevision', on_delete=models.CASCADE)
    path = models.TextField()
    order = models.PositiveIntegerField()

    class Meta:
        abstract = True


class SegmentPageLocation(BasePageLocation):
    segment = models.ForeignKey(Segment, on_delete=models.CASCADE, related_name='page_locations')

    @classmethod
    def from_segment_value(cls, page_revision, language, segment_value):
        segment = Segment.from_text(language, segment_value.html_with_ids)

        segment_page_loc, created = cls.objects.get_or_create(
            page_revision_id=pk(page_revision),
            path=segment_value.path,
            order=segment_value.order,
            segment=segment,
        )

        return segment_page_loc


class TemplatePageLocation(BasePageLocation):
    template = models.ForeignKey(Template, on_delete=models.CASCADE, related_name='page_locations')

    @classmethod
    def from_template_value(cls, page_revision, template_value):
        template = Template.from_template_value(template_value)

        template_page_loc, created = cls.objects.get_or_create(
            page_revision_id=pk(page_revision),
            path=template_value.path,
            order=template_value.order,
            template=template,
        )

        return template_page_loc



from django import forms
from django.conf.urls import url
from django.contrib import messages
from django.db import transaction
from django.utils.translation import ugettext_lazy
from django.views.decorators.http import require_POST
from django.views.generic import DetailView, ListView
from django.shortcuts import get_object_or_404, redirect

from wagtail.admin.views import generic
from wagtail.admin.viewsets.base import ViewSet
from wagtail.core.permission_policies import ModelPermissionPolicy

from ..action_modules import get_action_modules
from ..models import TranslationRequest


class TranslationRequestForm(forms.ModelForm):

    class Meta:
        model = TranslationRequest
        fields = ['source_locale']


class TranslationRequestListView(ListView):
    template_name = 'wagtail_localize_workflow/management/list.html'
    page_title = ugettext_lazy("Translaton requests")
    context_object_name = 'translation_requests'
    permission_policy = None
    list_url_name = None
    detail_url_name = None
    header_icon = ''


class TranslationRequestDetailView(DetailView):
    success_message = ugettext_lazy("Translation request '{0}' updated.")
    error_message = ugettext_lazy("The translation request could not be saved due to errors.")
    context_object_name = 'translation_request'
    template_name = 'wagtail_localize_workflow/management/detail.html'
    permission_policy = None
    list_url_name = None
    detail_url_name = None
    copy_for_translation_url_name = None
    header_icon = ''

    def get_context_data(self, object):
        context = super().get_context_data(object=object)
        pages = list(object.pages.order_by('id'))
        pages_by_id = {
            page.id: page
            for page in pages
        }

        # Add depths to pages
        for page in pages:
            # Pages are in depth-first-search order so parents are processed before their children
            if page.parent_id:
                page.depth = pages_by_id[page.parent_id].depth + 1
            else:
                page.depth = 0

        context['pages'] = pages

        context['action_modules'] = [
            action_module(self.request, object)
            for action_module in get_action_modules()
        ]

        # Hide action modules if their is_shown method returns false
        context['action_modules'] = [
            action_module for action_module in context['action_modules']
            if action_module.is_shown()
        ]

        return context


@require_POST
def copy_pages(request, translation_request_id):
    translation_request = get_object_or_404(TranslationRequest, id=translation_request_id)
    num_copied = 0

    for page in translation_request.pages.filter(is_completed=False):
        instance = page.source_revision.as_page_object()

        if instance.has_translation(translation_request.target_locale):
            continue

        with transaction.atomic():
            new_page = instance.copy_for_translation(translation_request.target_locale)

            # Update translation request
            page.is_completed = True
            page.completed_revision = new_page.get_latest_revision()
            page.save(update_fields=['is_completed', 'completed_revision'])

        num_copied += 1

    # TODO: Plural
    messages.success(request, "%d pages successfully copied" % num_copied)

    return redirect('wagtail_localize_workflow_management:detail', translation_request_id)


class CopyForTranslationView(DetailView):
    success_message = ugettext_lazy("Translation request '{0}' updated.")
    error_message = ugettext_lazy("The translation request could not be saved due to errors.")
    context_object_name = 'translation_request_page'
    template_name = 'wagtail_localize_workflow/management/copy_for_translation.html'
    permission_policy = None
    list_url_name = None
    detail_url_name = None
    copy_for_translation_url_name = None
    header_icon = ''

    def get_object(self):
        translation_request = super().get_object()
        return get_object_or_404(translation_request.pages.all(), id=self.kwargs['page_id'])

    def post(self, request, *args, **kwargs):
        translation_request_page = self.get_object()

        new_page = translation_request_page.source_page.specific.copy_for_translation(translation_request_page.request.target_locale)
        translation_request_page.is_completed = True
        translation_request_page.completed_revision = new_page.get_latest_revision()
        translation_request_page.save(update_fields=['is_completed', 'completed_revision'])
        return redirect('wagtailadmin_pages:edit', new_page.id)


class TranslationRequestViewSet(ViewSet):
    icon = 'site'

    model = TranslationRequest
    permission_policy = ModelPermissionPolicy(TranslationRequest)

    list_view_class = TranslationRequestListView
    detail_view_class = TranslationRequestDetailView
    copy_for_translation_view_class = CopyForTranslationView

    @property
    def list_view(self):
        return self.list_view_class.as_view(
            model=self.model,
            permission_policy=self.permission_policy,
            list_url_name=self.get_url_name('list'),
            detail_url_name=self.get_url_name('detail'),
            header_icon=self.icon,
        )

    @property
    def detail_view(self):
        return self.detail_view_class.as_view(
            model=self.model,
            permission_policy=self.permission_policy,
            list_url_name=self.get_url_name('list'),
            detail_url_name=self.get_url_name('detail'),
            copy_for_translation_url_name=self.get_url_name('copy_for_translation'),
            header_icon=self.icon,
        )

    @property
    def copy_for_translation_view(self):
        return self.copy_for_translation_view_class.as_view(
            model=self.model,
            permission_policy=self.permission_policy,
            list_url_name=self.get_url_name('list'),
            detail_url_name=self.get_url_name('detail'),
            copy_for_translation_url_name=self.get_url_name('copy_for_translation'),
            header_icon=self.icon,
        )

    def get_urlpatterns(self):
        return super().get_urlpatterns() + [
            url(r'^$', self.list_view, name='list'),
            url(r'^(?P<pk>\d+)/$', self.detail_view, name='detail'),
            url(r'^(\d+)/copy_pages/$', copy_pages, name='copy_pages'),
            url(r'^(?P<pk>\d+)/copy_for_translation/(?P<page_id>\d+)/$', self.copy_for_translation_view, name='copy_for_translation'),
        ]



# Generated by Django 2.1.10 on 2019-08-05 11:51

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_localize_workflow', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='translationrequest',
            name='source_locale',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='wagtail_localize.Locale'),
        ),
        migrations.AlterField(
            model_name='translationrequest',
            name='target_locale',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='wagtail_localize.Locale'),
        ),
        migrations.AlterField(
            model_name='translationrequest',
            name='target_root',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.Page'),
        ),
        migrations.AlterField(
            model_name='translationrequestpage',
            name='source_revision',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailcore.PageRevision'),
        ),
    ]



from django import forms

from django.contrib import messages
from django.db import transaction
from django.http import Http404
from django.shortcuts import get_object_or_404, render, redirect
from django.utils import timezone

from wagtail.core.models import Page

from wagtail_localize.models import get_translatable_models, Locale, TranslatablePageMixin

from ..models import TranslationRequest, TranslationRequestPage


class CreateTranslationRequestForm(forms.Form):
    locales = forms.ModelMultipleChoiceField(queryset=Locale.objects.none(), widget=forms.CheckboxSelectMultiple)
    include_subtree = forms.BooleanField(required=False)

    def __init__(self, page, *args, **kwargs):
        super().__init__(*args, **kwargs)

        page_descendant_count = page.get_descendants().type(tuple(get_translatable_models())).count()
        if page_descendant_count > 0:
            self.fields['include_subtree'].help_text = f"This will add {page_descendant_count} additional pages to the request"
        else:
            self.fields['include_subtree'].widget = forms.HiddenInput()

        self.fields['locales'].queryset = Locale.objects.filter(is_active=True).exclude(id=page.locale_id)


def create_translation_request(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    if not isinstance(page, TranslatablePageMixin):
        raise Http404

    if request.method == 'POST':
        form = CreateTranslationRequestForm(page, request.POST)

        if form.is_valid():
            with transaction.atomic():
                source_locale = page.locale
                revision = page.get_latest_revision()
                translatable_models = tuple(get_translatable_models())

                # If the page doesn't have a revision (eg, came from import-export), create one.
                if revision is None:
                    revision = page.save_revision(changed=False)

                for target_locale in form.cleaned_data['locales']:
                    # Work out target root. While we're at it, get list of any translatable
                    # ancestor pages that don't have translations yet as these will need to
                    # be translated too.
                    required_ancestors = []
                    current_page = page
                    target_root = current_page.get_translation_or_none(target_locale)
                    while target_root is None:
                        current_page = current_page.get_parent()

                        if issubclass(current_page.specific_class, translatable_models):
                            target_root = current_page.specific.get_translation_or_none(target_locale)

                            if target_root is None:
                                required_ancestors.append(current_page)
                        else:
                            target_root = current_page
                            break

                    # Create translation request
                    translation_request = TranslationRequest.objects.create(
                        source_locale=source_locale,
                        target_locale=target_locale,
                        target_root=target_root,
                        created_at=timezone.now(),
                        created_by=request.user,
                    )

                    # Add ancestor pages to translation request
                    parent_item = None
                    for ancestor_page in required_ancestors:
                        source_revision = ancestor_page.get_latest_revision()
                        if source_revision is None:
                            source_revision = ancestor_page.specific.save_revision(changed=False)

                        parent_item = TranslationRequestPage.objects.create(
                            request=translation_request,
                            source_revision=source_revision,
                            parent=parent_item,
                        )

                    # Now add the requested page
                    parent_item = TranslationRequestPage.objects.create(
                        request=translation_request,
                        source_revision=revision,
                        parent=parent_item,
                    )

                    # Now add the sub tree
                    if form.cleaned_data['include_subtree']:
                        def _walk(current_page, parent_item):
                            for child_page in current_page.get_children():
                                if not issubclass(child_page.specific_class, translatable_models):
                                    continue

                                source_revision = child_page.get_latest_revision()
                                if source_revision is None:
                                    source_revision = child_page.specific.save_revision(changed=False)

                                child_item = TranslationRequestPage.objects.create(
                                    request=translation_request,
                                    source_revision=source_revision,
                                    parent=parent_item,
                                )

                                if child_page.numchild:
                                    _walk(child_page, child_item)

                        _walk(page, parent_item)

                messages.success(request, "The translation request was submitted successfully")
                return redirect('wagtailadmin_explore', page.get_parent().id)
    else:
        form = CreateTranslationRequestForm(page)

    return render(request, 'wagtail_localize_workflow/create_translation_request.html', {
        'page': page,
        'form': form,
    })



from django.apps import AppConfig


class WagtailLocalizeWorkflowAppConfig(AppConfig):
    label = 'wagtail_localize_workflow'
    name = 'wagtail_localize.admin.workflow'
    verbose_name = "Wagtail Localize Workflow"



import functools

from django.template.loader import render_to_string

from wagtail.core import hooks


class BaseActionModule:
    template_name = None

    def __init__(self, request, translation_request):
        self.request = request
        self.translation_request = translation_request

    def is_shown(self):
        return True

    def render(self):
        context = {
            'translation_request': self.translation_request,
        }

        return render_to_string(self.template_name, context, request=self.request)


class CopyPagesActionModule(BaseActionModule):
    template_name = 'wagtail_localize_workflow/action_modules/copy_pages.html'


@functools.lru_cache()
def get_action_modules():
    action_modules = [
        CopyPagesActionModule,
    ]

    for fn in hooks.get_hooks('wagtail_localize_workflow_register_action_modules'):
        new_action_modules = fn()
        if new_action_modules:
            action_modules.extend(new_action_modules)

    return action_modules



# Generated by Django 2.1.7 on 2019-03-25 17:34

from django.conf import settings
from django.db import migrations

from wagtail_localize.compat import get_supported_language_variant


def initial_data(apps, schema_editor):
    Language = apps.get_model('wagtail_localize.Language')
    Region = apps.get_model('wagtail_localize.Region')
    Locale = apps.get_model('wagtail_localize.Locale')

    default_language, created = Language.objects.get_or_create(
        code=get_supported_language_variant(settings.LANGUAGE_CODE),
    )

    default_region = Region.objects.create(
        name='Default',
        slug='default',
        is_default=True,
    )

    default_region.languages.add(default_language)

    Locale.objects.create(
        language=default_language,
        region=default_region,
        is_active=True,
    )


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_localize', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(initial_data, migrations.RunPython.noop)
    ]



from django.test import TestCase

from wagtail_localize.segments import SegmentValue
from wagtail_localize.segments.html import extract_html_segments, extract_html_elements, restore_html_elements


class TestExtractHTMLSegment(TestCase):
    def test_extract_segments(self):
        template, segments = extract_html_segments(
            """
            <p><b>Bread</b>\xa0is a\xa0<a href="https://en.wikipedia.org/wiki/Staple_food">staple food</a>\xa0prepared from a\xa0<a href="https://en.wikipedia.org/wiki/Dough">dough</a>\xa0of\xa0<a href="https://en.wikipedia.org/wiki/Flour">flour</a>\xa0and\xa0<a href="https://en.wikipedia.org/wiki/Water">water</a>, usually by\xa0<a href="https://en.wikipedia.org/wiki/Baking">baking</a>. Throughout recorded history it has been popular around the world and is one of the oldest artificial foods, having been of importance since the dawn of\xa0<a href="https://en.wikipedia.org/wiki/Agriculture#History">agriculture</a>.</p>
            <p>Proportions of types of flour and other ingredients vary widely, as do modes of preparation. As a result, types, shapes, sizes, and textures of breads differ around the world. Bread may be\xa0<a href="https://en.wikipedia.org/wiki/Leaven">leavened</a>\xa0by processes such as reliance on naturally occurring\xa0<a href="https://en.wikipedia.org/wiki/Sourdough">sourdough</a>\xa0microbes, chemicals, industrially produced yeast, or high-pressure aeration. Some bread is cooked before it can leaven, including for traditional or religious reasons. Non-cereal ingredients such as fruits, nuts and fats may be included. Commercial bread commonly contains additives to improve flavor, texture, color, shelf life, and ease of manufacturing.</p>
            """
        )

        self.assertHTMLEqual(template,
            """
            <p><text position="0"></text></p>
            <p><text position="1"></text></p>
            """
        )

        self.assertEqual(segments, [
            '<b>Bread</b>\xa0is a\xa0<a href="https://en.wikipedia.org/wiki/Staple_food">staple food</a>\xa0prepared from a\xa0<a href="https://en.wikipedia.org/wiki/Dough">dough</a>\xa0of\xa0<a href="https://en.wikipedia.org/wiki/Flour">flour</a>\xa0and\xa0<a href="https://en.wikipedia.org/wiki/Water">water</a>, usually by\xa0<a href="https://en.wikipedia.org/wiki/Baking">baking</a>. Throughout recorded history it has been popular around the world and is one of the oldest artificial foods, having been of importance since the dawn of\xa0<a href="https://en.wikipedia.org/wiki/Agriculture#History">agriculture</a>.',
            'Proportions of types of flour and other ingredients vary widely, as do modes of preparation. As a result, types, shapes, sizes, and textures of breads differ around the world. Bread may be\xa0<a href="https://en.wikipedia.org/wiki/Leaven">leavened</a>\xa0by processes such as reliance on naturally occurring\xa0<a href="https://en.wikipedia.org/wiki/Sourdough">sourdough</a>\xa0microbes, chemicals, industrially produced yeast, or high-pressure aeration. Some bread is cooked before it can leaven, including for traditional or religious reasons. Non-cereal ingredients such as fruits, nuts and fats may be included. Commercial bread commonly contains additives to improve flavor, texture, color, shelf life, and ease of manufacturing.',
        ])

    def test_extract_segments_2(self):
        template, segments = extract_html_segments(
            """
            <h1>Foo bar baz</h1>
            <p>This is a paragraph. <b>This is some bold <i>and now italic</i></b> text</p>
            <p>&lt;script&gt; this should be interpreted as text.</p>
            <ul>
                <li>List item one</li>
                <li><b>List item two</li>
            </ul>
            <img src="foo" alt="This bit isn't translatable">
            """
        )

        self.assertHTMLEqual(template,
            """
            <h1><text position="0"></text></h1>
            <p><text position="1"></text></p>
            <p><text position="2"></text></p>
            <ul>
                <li><text position="3"></text></li>
                <li><text position="4"></text></li>
            </ul>
            <img alt="This bit isn\'t translatable" src="foo">
            """
        )

        self.assertEqual(segments, [
            'Foo bar baz',
            'This is a paragraph. <b>This is some bold <i>and now italic</i></b> text',
            '&lt;script&gt; this should be interpreted as text.',
            'List item one',
            '<b>List item two</b>'
        ])


class TestExtractHTMLElements(TestCase):
    def test_extract_html_elements(self):
        text, elements = extract_html_elements("This is a paragraph. <b>This is some bold <i>and now italic</i></b> text")

        self.assertEqual(text, "This is a paragraph. This is some bold and now italic text")
        self.assertEqual(elements, [(39, 53, 'i', {}), (21, 53, 'b', {})])

    def test_special_chars_unescaped(self):
        text, elements = extract_html_elements("<b>foo</b><i> &amp; bar</i>")

        self.assertEqual(text, "foo & bar")
        self.assertEqual(elements, [(0, 3, 'b', {}), (3, 9, 'i', {})])


class TestRestoreHTMLElements(TestCase):
    def test_restore_html_elements(self):
        html = restore_html_elements("This is a paragraph. This is some bold and now italic text", [(39, 53, 'i', {}), (21, 53, 'b', {})])

        self.assertEqual(html, "This is a paragraph. <b>This is some bold <i>and now italic</i></b> text")

    def test_special_chars_escaped(self):
        html = restore_html_elements("foo & bar", [(0, 3, 'b', {}), (3, 9, 'i', {})])

        self.assertEqual(html, "<b>foo</b><i> &amp; bar</i>")


class TestHTMLSegmentValue(TestCase):
    def test_text(self):
        segment = SegmentValue.from_html("foo", "This is a paragraph. <b>This is some bold <i>and now italic</i></b> text")
        self.assertEqual(segment.text, "This is a paragraph. This is some bold and now italic text")

    def test_html(self):
        segment = SegmentValue.from_html("foo", "This is a paragraph. <b>This is some bold <i>and now italic</i></b> text")
        self.assertEqual(segment.html, "This is a paragraph. <b>This is some bold <i>and now italic</i></b> text")

    def test_html_with_ids(self):
        segment = SegmentValue.from_html("foo", '<b>Bread</b>\xa0is a\xa0<a href="https://en.wikipedia.org/wiki/Staple_food">staple food</a>\xa0prepared from a\xa0<a href="https://en.wikipedia.org/wiki/Dough">dough</a>\xa0of\xa0<a href="https://en.wikipedia.org/wiki/Flour">flour</a>\xa0and\xa0<a href="https://en.wikipedia.org/wiki/Water">water</a>, usually by\xa0<a href="https://en.wikipedia.org/wiki/Baking">baking</a>. Throughout recorded history it has been popular around the world and is one of the oldest artificial foods, having been of importance since the dawn of\xa0<a href="https://en.wikipedia.org/wiki/Agriculture#History">agriculture</a>.')
        self.assertEqual(segment.html_with_ids, '<b>Bread</b> is a <a id="a1">staple food</a> prepared from a <a id="a2">dough</a> of <a id="a3">flour</a> and <a id="a4">water</a>, usually by <a id="a5">baking</a>. Throughout recorded history it has been popular around the world and is one of the oldest artificial foods, having been of importance since the dawn of <a id="a6">agriculture</a>.')



from collections import OrderedDict

import polib
from django.core.management.base import BaseCommand, CommandError
from django.contrib.contenttypes.models import ContentType
from wagtail.core.models import Page

from wagtail_localize.models import get_translatable_models, Language
from wagtail_localize.translation_memory.models import Segment, SegmentTranslation, SegmentPageLocation, TemplatePageLocation
from wagtail_localize.translation_memory.utils import insert_segments
from wagtail_localize.segments import TemplateValue
from wagtail_localize.segments.extract import extract_segments


class Command(BaseCommand):

    def handle(self, **options):
        src_lang = Language.default()
        tgt_lang = Language.objects.filter(code='fr').first()
        messages = OrderedDict()

        def get_page_revision(page):
            if page.live_revision:
                return page.live_revision

            latest_revision = page.get_latest_revision()
            if latest_revision:
                return latest_revision

            return page.save_revision(changed=False)

        for model in get_translatable_models():
            if not issubclass(model, Page):
                continue

            if model._meta.abstract:
                continue

            content_type = ContentType.objects.get_for_model(model)
            pages = model.objects.live().filter(content_type=content_type, locale=src_lang)

            for page in pages:
                segments = extract_segments(page)
                insert_segments(get_page_revision(page), src_lang, segments)
                for segment in segments:
                    if not isinstance(segment, TemplateValue):
                        text = segment.html
                        if text not in messages:
                            messages[text] = []

                        messages[text].append((page.url_path, segment.path))

        po = polib.POFile()
        po.metadata = {
            'Project-Id-Version': '1.0',
            'Report-Msgid-Bugs-To': 'you@example.com',
            'POT-Creation-Date': '2007-10-18 14:00+0100',
            'PO-Revision-Date': '2007-10-18 14:00+0100',
            'Last-Translator': 'you <you@example.com>',
            'Language-Team': 'English <yourteam@example.com>',
            'MIME-Version': '1.0',
            'Content-Type': 'text/plain; charset=utf-8',
            'Content-Transfer-Encoding': '8bit',
        }

        for segment, occurances in messages.items():
            existing_translation = ''
            if tgt_lang:
                translation = SegmentTranslation.objects.filter(translation_of__text=segment, translation_of__locale=src_lang, locale=tgt_lang).first()

                if translation:
                    existing_translation = translation.text

            po.append(
                polib.POEntry(
                    msgid=segment,
                    msgstr=existing_translation,
                    occurrences=occurances
                )
            )

        po.save('site.po')



from django.db import models

from wagtail.core import blocks
from wagtail.core.fields import RichTextField, StreamField
from wagtail.core.rich_text import RichText
from wagtail.images.blocks import ImageChooserBlock
from wagtail.snippets.blocks import SnippetChooserBlock

from wagtail_localize.models import TranslatableMixin
from wagtail_localize.segments import SegmentValue, TemplateValue

from .html import extract_html_segments


class StreamFieldSegmentExtractor:
    def __init__(self, field):
        self.field = field

    def handle_block(self, block_type, block_value):
        if hasattr(block_type, 'get_translatable_segments'):
            return block_type.get_translatable_segments(block_value)

        elif isinstance(block_type, (blocks.CharBlock, blocks.TextBlock)):
            return [SegmentValue('', block_value)]

        elif isinstance(block_type, blocks.RichTextBlock):
            template, texts = extract_html_segments(block_value.source)

            return [
                TemplateValue('', 'html', template, len(texts))
            ] + [
                SegmentValue.from_html(str(position), text)
                for position, text in enumerate(texts)
            ]

        elif isinstance(block_type, (ImageChooserBlock, SnippetChooserBlock)):
            return self.handle_related_object_block(block_value)

        elif isinstance(block_type, blocks.StructBlock):
            return self.handle_struct_block(block_value)

        elif isinstance(block_type, blocks.ListBlock):
            return self.handle_list_block(block_value)

        elif isinstance(block_type, blocks.ChoiceBlock):
            return []

        else:
            raise Exception("Unrecognised StreamField block type '{}'. Have you implemented get_translatable_segments on this class?".format(block_type.__class__.__name__))

    def handle_related_object_block(self, related_object):
        if related_object is None or not isinstance(related_object, TranslatableMixin):
            return []

        return extract_segments(related_object)

    def handle_struct_block(self, struct_block):
        segments = []

        for field_name, block_value in struct_block.items():
            block_type = struct_block.block.child_blocks[field_name]
            segments.extend(segment.wrap(field_name) for segment in self.handle_block(block_type, block_value))

        return segments

    def handle_list_block(self, list_block):
        # TODO
        return []

    def handle_stream_block(self, stream_block):
        segments = []

        for block in stream_block:
            segments.extend(segment.wrap(block.id) for segment in self.handle_block(block.block, block.value))

        return segments


def extract_segments(instance):
    segments = []

    for field in instance.get_translatable_fields():
        if hasattr(field, 'get_translatable_segments'):
            segments.extend(segment.wrap(field.name) for segment in field.get_translatable_segments(field.value_from_object(instance)))

        elif isinstance(field, StreamField):
            segments.extend(segment.wrap(field.name) for segment in StreamFieldSegmentExtractor(field).handle_stream_block(field.value_from_object(instance)))

        elif isinstance(field, RichTextField):
            template, texts = extract_html_segments(field.value_from_object(instance))

            field_segments = [
                TemplateValue('', 'html', template, len(texts))
            ] + [
                SegmentValue.from_html(str(position), text)
                for position, text in enumerate(texts)
            ]

            segments.extend(segment.wrap(field.name) for segment in field_segments)

        elif isinstance(field, (models.TextField, models.CharField)):
            if not field.choices:
                segments.append(SegmentValue(field.name, field.value_from_object(instance)))

        elif isinstance(field, (models.ForeignKey)) and issubclass(field.related_model, TranslatableMixin):
            related_instance = getattr(instance, field.name)

            if related_instance:
                segments.extend(segment.wrap(field.name) for segment in extract_segments(related_instance))

        elif isinstance(field, (models.ManyToOneRel)) and issubclass(field.related_model, TranslatableMixin):
            manager = getattr(instance, field.name)

            for child_instance in manager.all():
                segments.extend(segment.wrap(f'{field.name}.{child_instance.translation_key}') for segment in extract_segments(child_instance))


    class Counter:
        def __init__(self):
            self.value = 0

        def next(self):
            self.value += 1
            return self.value

    counter = Counter()

    return [segment.with_order(counter.next()) for segment in segments if not segment.is_empty()]



# Generated by Django 2.1.10 on 2019-08-13 15:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wagtail_localize_translation_memory', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='segmentpagelocation',
            name='order',
            field=models.PositiveIntegerField(default=0),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='templatepagelocation',
            name='order',
            field=models.PositiveIntegerField(default=0),
            preserve_default=False,
        ),
    ]



import os
import uuid

from django.apps import apps
from django.conf import settings
from django.db import models, transaction
from django.db.models import Exists, OuterRef, Q
from django.db.models.signals import post_save, m2m_changed
from django.dispatch import receiver
from django.utils import translation
from wagtail.core.models import Page
from wagtail.images.models import AbstractImage

from .compat import get_languages, get_supported_language_variant
from .utils import find_available_slug


class LanguageManager(models.Manager):
    use_in_migrations = True

    def default(self):
        default_code = get_supported_language_variant(settings.LANGUAGE_CODE)
        return self.get(code=default_code)

    def default_id(self):
        return self.default().id


class Language(models.Model):
    code = models.CharField(max_length=100, unique=True)
    is_active = models.BooleanField(default=True)

    objects = LanguageManager()

    class Meta:
        ordering = ['-is_active', 'code']

    @classmethod
    def get_active(cls):
        default_code = get_supported_language_variant(translation.get_language())

        language, created = cls.objects.get_or_create(
            code=default_code,
        )

        return language

    def as_rfc5646_language_tag(self):
        """
        Returns the language code using the capitalisation as specified in RFC5646.

        For example: en-GB (not en-gb or en_GB)

        https://tools.ietf.org/html/rfc5646
        """
        # These are all the script codes used in Django's default LANGUAGES
        # They exist instead of a country, but they need to be capitalised differently.
        script_codes = ['latn', 'hans', 'hant']

        components = self.code.split('-')

        if len(components) == 1:
            # en
            return self.code.lower()
        elif len(components) == 2:
            # en-gb or zh-hans

            if components[1] in script_codes:
                # zh-hans => zh-Hans
                return components[0].lower() + '-' + components[1].title()
            else:
                # en-gb => en-GB
                return components[0].lower() + '-' + components[1].title()
        else:
            # Not sure what to do. Returning the existing code is the
            return self.code

    @classmethod
    def get_by_rfc5646_language_tag(cls, language_tag):
        return cls.objects.get(code=language_tag.lower())

    def get_display_name(self):
        return get_languages().get(self.code)

    def __str__(self):
        display_name = self.get_display_name()

        if display_name:
            return f'{display_name} ({self.code})'
        else:
            return self.code


class RegionManager(models.Manager):
    use_in_migrations = True

    def default(self):
        return self.filter(is_default=True).first()

    def default_id(self):
        default = self.default()

        if default:
            return default.id


class Region(models.Model):
    name = models.CharField(max_length=100)
    slug = models.SlugField(max_length=100, unique=True)
    languages = models.ManyToManyField(Language)
    is_default = models.BooleanField(default=False)
    is_active = models.BooleanField(default=True)

    objects = RegionManager()

    class Meta:
        ordering = ['-is_active', '-is_default', 'name']

    def __str__(self):
        return self.name


# Add new languages to default region automatically
# This allows sites that don't care about regions to still work
@receiver(post_save, sender=Language)
def add_new_languages_to_default_region(sender, instance, created, **kwargs):
    if created:
        default_region = Region.objects.default()

        if default_region is not None:
            default_region.languages.add(instance)


class LocaleManager(models.Manager):
    use_in_migrations = True

    def default(self):
        locale, created = self.get_or_create(
            region_id=Region.objects.default_id(),
            language_id=Language.objects.default_id(),
            is_active=True,
        )

        return locale

    def default_id(self):
        return self.default().id


# A locale gives an individual record to a region/language combination
# I prefer this way as it allows you to easily reorganise your regions and languages
# after there is already content entered.
# Note: these are managed entirely through signal handlers so don't update them directly
# without also updating the Language/Region models as well.
class Locale(models.Model):
    region = models.ForeignKey(Region, on_delete=models.CASCADE, related_name='locales')
    language = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='locales')
    is_active = models.BooleanField(default=True)

    objects = LocaleManager()

    class Meta:
        unique_together = [
            ('region', 'language'),
        ]
        ordering = ['-is_active', '-region__is_default', 'region__name', 'language__code']

    # For backwards compatibility
    @classmethod
    def default(cls):
        return cls.objects.default()

    # For backwards compatibility
    @classmethod
    def default_id(cls):
        return cls.objects.default_id()

    def __str__(self):
        return f"{self.region.name} / {self.language.get_display_name()}"

    @property
    def slug(self):
        slug = self.language.code

        if self.region != Region.objects.default():
            return f'{self.region.slug}-{self.language.code}'
        else:
            return self.language.code

    def get_all_pages(self):
        """
        Returns a queryset of all pages that have been translated into this locale.
        """
        q = Q()

        for model in get_translatable_models():
            q |= Q(id__in=model.objects.filter(locale=self).values_list('id', flat=True))

        return Page.objects.filter(q)


# Update Locale.is_active when Language.is_active is changed
@receiver(post_save, sender=Language)
def update_locales_on_language_change(sender, instance, **kwargs):
    if instance.is_active:
        # Activate locales iff the language is selected on the region
        (
            Locale.objects
            .annotate(
                is_active_on_region=Exists(
                    Region.languages.through.objects.filter(
                        region_id=OuterRef('region_id'),
                        language_id=instance.id,
                    )
                )
            )
            .filter(
                language=instance,
                region__is_active=True,
                is_active_on_region=True,
                is_active=False,
            )
            .update(is_active=True)
        )

    else:
        # Deactivate locales with this language
        Locale.objects.filter(language=instance, is_active=True).update(is_active=False)


# Update Locale.is_active when Region.is_active is changed
@receiver(post_save, sender=Region)
def update_locales_on_region_change(sender, instance, **kwargs):
    if instance.is_active:
        # Activate locales with this region
        (
            Locale.objects
            .annotate(
                is_active_on_region=Exists(
                    Region.languages.through.objects.filter(
                        region_id=instance.id,
                        language_id=OuterRef('language_id'),
                    )
                )
            )
            .filter(
                region=instance,
                is_active_on_region=True,
                is_active=False,
            )
            .update(is_active=True)
        )

    else:
        # Deactivate locales with this region
        Locale.objects.filter(region=instance, is_active=True).update(is_active=False)


# Add/remove locales when languages are added/removed from regions
@receiver(m2m_changed, sender=Region.languages.through)
def update_locales_on_region_languages_change(sender, instance, action, pk_set, **kwargs):
    if action == 'post_add':
        for language_id in pk_set:
            Locale.objects.update_or_create(
                region=instance,
                language_id=language_id,
                defaults={
                    # Note: only activate locale if language is active
                    'is_active': Language.objects.filter(id=language_id, is_active=True).exists(),
                }
            )
    elif action == 'post_remove':
        for language_id in pk_set:
            Locale.objects.update_or_create(
                region=instance,
                language_id=language_id,
                defaults={
                    'is_active': False,
                }
            )


def default_locale_id():
    return Locale.objects.default_id()


class TranslatableMixin(models.Model):
    translation_key = models.UUIDField(default=uuid.uuid4, editable=False)
    locale = models.ForeignKey(Locale, on_delete=models.PROTECT, related_name='+', default=default_locale_id)

    translatable_fields = []

    def get_translations(self, inclusive=False):
        translations = self.__class__.objects.filter(translation_key=self.translation_key)

        if inclusive is False:
            translations = translations.exclude(id=self.id)

        return translations

    def get_translation(self, locale):
        return self.get_translations(inclusive=True).get(locale=locale)

    def get_translation_or_none(self, locale):
        try:
            return self.get_translation(locale)
        except self.__class__.DoesNotExist:
            return None

    def has_translation(self, locale):
        return self.get_translations(inclusive=True).filter(locale=locale).exists()

    def copy_for_translation(self, locale):
        """
        Copies this instance for the specified locale.
        """
        translated = self.__class__.objects.get(id=self.id)
        translated.id = None
        translated.locale = locale

        if isinstance(self, AbstractImage):
            # As we've copied the image record we also need to copy the original image file itself.
            # This is in case either image record is changed or deleted, the other record will still
            # have its file.
            file_stem, file_ext = os.path.splitext(self.file.name)
            new_name = f'{file_stem}-{locale.slug}{file_ext}'
            translated.file = ContentFile(self.file.read(), name=new_name)

        return translated

    @classmethod
    def get_translation_model(self):
        """
        Gets the model which manages the translations for this model.
        (The model that has the "translation_key" and "locale" fields)
        Most of the time this would be the current model, but some sites
        may have intermediate concrete models between wagtailcore.Page and
        the specfic page model.
        """
        return self._meta.get_field('locale').model

    @classmethod
    def get_translatable_fields(cls):
        for field in cls._meta.get_fields():
            if field.name in cls.translatable_fields:
                yield field

    class Meta:
        abstract = True
        unique_together = [
            ('translation_key', 'locale'),
        ]


class TranslatablePageMixin(TranslatableMixin):

    def copy(self, reset_translation_key=True, **kwargs):
        # If this is a regular copy (not copy_for_translation) we should change the translation_key
        # of the page and all child objects so they don't clash with the original page

        if reset_translation_key:
            if 'update_attrs' in kwargs:
                if 'translation_key' not in kwargs['update_attrs']:
                    kwargs['update_attrs']['translation_key'] = uuid.uuid4()

            else:
                kwargs['update_attrs'] = {
                    'translation_key': uuid.uuid4(),
                }

            original_process_child_object = kwargs.pop('process_child_object')

            def process_child_object(original_page, page_copy, child_relation, child_object):
                # Change translation keys of translatable child objects
                if isinstance(child_object, TranslatableMixin):
                    child_object.translation_key = uuid.uuid4()

                if original_process_child_object is not None:
                    original_process_child_object(original_page, page_copy, child_relation, child_object)

            kwargs['process_child_object'] = process_child_object

        return super().copy(**kwargs)

    @transaction.atomic
    def copy_for_translation(self, locale, copy_parents=False, exclude_fields=None):
        """
        Copies this page for the specified locale.
        """
        # Find the translated version of the parent page to create the new page under
        parent = self.get_parent().specific
        slug = self.slug
        if isinstance(parent, TranslatablePageMixin):
            try:
                translated_parent = parent.get_translation(locale)
            except parent.__class__.DoesNotExist:
                if not copy_parents:
                    return

                translated_parent = parent.copy_for_translation(locale, copy_parents=True)
        else:
            translated_parent = parent

            # Append locale code to slug as the new page
            # will be created in the same section as the existing one
            slug += '-' + locale.slug

        # Find available slug for new page
        slug = find_available_slug(translated_parent, slug)

        # Update locale on translatable child objects as well
        def process_child_object(original_page, page_copy, child_relation, child_object):
            if isinstance(child_object, TranslatableMixin):
                child_object.locale = locale

        return self.copy(
            to=translated_parent,
            update_attrs={'locale': locale, 'slug': slug},
            copy_revisions=False,
            keep_live=False,
            reset_translation_key=False,
            process_child_object=process_child_object,
            exclude_fields=exclude_fields,
        )

    def with_content_json(self, content_json):
        page = super().with_content_json(content_json)
        page.translation_key = self.translation_key
        page.locale = self.locale
        return page

    def get_translation_for_request(self, request):
        """
        Returns the translation of this page that should be used to route
        the specified request.
        """
        language_code = translation.get_supported_language_variant(request.LANGUAGE_CODE)

        try:
            locale = Locale.objects.get(language__code=language_code, region__is_default=True)
            return self.get_translation(locale)

        except Locale.DoesNotExist:
            return

        except Page.DoesNotExist:
            return

    def route(self, request, path_components):
        # When this mixin is applied to a page, this override
        # will change the routing behaviour to route requests to
        # the correct translation based on the language code on
        # the request.

        # Check that an ancestor hasn't already routed the request into
        # a single-language section of the site.
        if not getattr(request, '_wagtail_localize_routed', False):
            # If the site root is translatable, reroute based on language
            page = self.get_translation_for_request(request)

            if page is not None:
                request._wagtail_localize_routed = True
                return page.route(request, path_components)

        return super().route(request, path_components)

    class Meta:
        abstract = True
        unique_together = [
            ('translation_key', 'locale'),
        ]


class BootstrapTranslatableMixin(TranslatableMixin):
    """
    A version of TranslatableMixin without uniqueness constraints.

    This is to make it easy to transition existing models to being translatable.

    The process is as follows:
     - Add BootstrapTranslatableMixin to the model
     - Run makemigrations
     - Create a data migration for each app, then use the BootstrapTranslatableModel operation in
       wagtail_localize.bootstrap on each model in that app
     - Change BootstrapTranslatableMixin to TranslatableMixin (or TranslatablePageMixin, if it's a page model)
     - Run makemigrations again
     - Migrate!
    """
    translation_key = models.UUIDField(null=True, editable=False)
    locale = models.ForeignKey(Locale, on_delete=models.PROTECT, null=True, related_name='+')

    class Meta:
        abstract = True


def get_translatable_models(include_subclasses=False):
    """
    Returns a list of all concrete models that inherit from TranslatableMixin.
    By default, this only includes models that are direct children of TranslatableMixin,
    to get all models, set the include_subclasses attribute to True.
    """
    translatable_models = [
        model for model in apps.get_models()
        if issubclass(model, TranslatableMixin) and not model._meta.abstract
    ]

    if include_subclasses is False:
        # Exclude models that inherit from another translatable model
        root_translatable_models = set()

        for model in translatable_models:
            root_translatable_models.add(model.get_translation_model())

        translatable_models = [
            model for model in translatable_models
            if model in root_translatable_models
        ]

    return translatable_models



from django.apps import AppConfig


class WagtailLocalizeSitesAppConfig(AppConfig):
    label = 'wagtail_localize_sites'
    name = 'wagtail_localize.experimental.sites'
    verbose_name = "Wagtail Localize Sites"



from collections import OrderedDict

import polib
from django.core.management.base import BaseCommand, CommandError
from django.contrib.contenttypes.models import ContentType
from wagtail.core.models import Page

from wagtail_localize.models import Language

from wagtail_localize.translation_memory.translation_ingestor import ingest_translations


class Command(BaseCommand):

    def handle(self, **options):
        src_lang = Language.default()
        tgt_lang = Language.objects.filter(code='fr').first()

        po = polib.pofile('site-fr.po')
        translations = [(message.msgid, message.msgstr) for message in po]
        ingest_translations(src_lang, tgt_lang, translations)



from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from wagtail_localize.models import Language


class Command(BaseCommand):
    help = 'Synchronises the value of settings.LANGUAGES with the Language model'

    def handle(self, **options):
        language_codes = dict(settings.LANGUAGES).keys()

        for language_code in language_codes:
            Language.objects.update_or_create(
                code=language_code,
                defaults={
                    'is_active': True,
                }
            )

        for deactivated_languages in Language.objects.exclude(code__in=language_codes).filter(is_active=True):
            deactivated_languages.is_active = False
            deactivated_languages.save()



#!/usr/bin/env python

import sys, os

from setuptools import setup, find_packages

# Hack to prevent "TypeError: 'NoneType' object is not callable" error
# in multiprocessing/util.py _exit_function when setup.py exits
# (see http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html)
try:
    import multiprocessing
except ImportError:
    pass


setup(
    name='wagtail-localize',
    version='0.2',
    description='Wagtail Internationalisation Framework',
    author='Karl Hobley',
    author_email='karl@kaed.uk',
    url='',
    packages=find_packages(),
    license='BSD',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=[],
    zip_safe=False,
)



from django.apps import AppConfig


class WagtailLocalizeLanguageSwitchAppConfig(AppConfig):
    label = 'wagtail_localize_language_switch'
    name = 'wagtail_localize.admin.language_switch'
    verbose_name = "Wagtail Localize Language switch"



from django.apps import AppConfig


class WagtailLocalizeGoogleTranslateAppConfig(AppConfig):
    label = 'wagtail_localize_google_translate'
    name = 'wagtail_localize.translation_engines.google_translate'
    verbose_name = "Wagtail Localize Google Translate translation engine"



import uuid

from django.conf import settings
from django.db import migrations

from .compat import get_supported_language_variant


def bootstrap_translatable_model(model, locale):
    """
    This function populates the "translation_key", and "locale" fields on model instances that were created
    before wagtail-localize was added to the site.

    This can be called from a data migration, or instead you could use the "boostrap_translatable_models"
    management command.
    """
    # TODO: Optimise for databases that have a UUID4 function
    for instance in model.objects.filter(translation_key__isnull=True).defer().iterator():
        instance.translation_key = uuid.uuid4()
        instance.locale = locale
        instance.save(update_fields=['translation_key', 'locale'])


class BootstrapTranslatableModel(migrations.RunPython):
    def __init__(self, model_string, language_code=None, region_slug='default'):
        def forwards(apps, schema_editor):
            model = apps.get_model(model_string)
            Locale = apps.get_model('wagtail_localize.Locale')

            if language_code is not None:
                locale = Locale.objects.get(language__code=language_code, region__slug=region_slug)
            else:
                locale = Locale.objects.default()

            bootstrap_translatable_model(model, locale)

        def backwards(apps, schema_editor):
            pass

        super().__init__(forwards, backwards)



# Generated by Django 2.1.9 on 2019-06-21 09:16

from django.db import migrations, models
import django.db.models.deletion
import wagtail_localize.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Language',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=100, unique=True)),
                ('is_active', models.BooleanField(default=True)),
            ],
            options={
                'ordering': ['-is_active', 'code'],
            },
            managers=[
                ('objects', wagtail_localize.models.LanguageManager()),
            ],
        ),
        migrations.CreateModel(
            name='Locale',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True)),
                ('language', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='locales', to='wagtail_localize.Language')),
            ],
            options={
                'ordering': ['-is_active', '-region__is_default', 'region__name', 'language__code'],
            },
            managers=[
                ('objects', wagtail_localize.models.LocaleManager()),
            ],
        ),
        migrations.CreateModel(
            name='Region',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('slug', models.SlugField(max_length=100, unique=True)),
                ('is_default', models.BooleanField(default=False)),
                ('is_active', models.BooleanField(default=True)),
                ('languages', models.ManyToManyField(to='wagtail_localize.Language')),
            ],
            options={
                'ordering': ['-is_active', '-is_default', 'name'],
            },
            managers=[
                ('objects', wagtail_localize.models.RegionManager()),
            ],
        ),
        migrations.AddField(
            model_name='locale',
            name='region',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='locales', to='wagtail_localize.Region'),
        ),
        migrations.AlterUniqueTogether(
            name='locale',
            unique_together={('region', 'language')},
        ),
    ]



import tempfile
from collections import defaultdict

import polib
from django.contrib import messages
from django.db import transaction
from django.http import HttpResponse
from django.views.decorators.http import require_GET, require_POST
from django.shortcuts import get_object_or_404, redirect
from django.utils import timezone
from django.utils.text import slugify

from wagtail_localize.admin.workflow.models import TranslationRequest
from wagtail_localize.segments import SegmentValue, TemplateValue
from wagtail_localize.segments.extract import extract_segments
from wagtail_localize.segments.ingest import ingest_segments


@require_GET
def download(request, translation_request_id):
    translation_request = get_object_or_404(TranslationRequest, id=translation_request_id)

    # Extract messages from pages
    messages = defaultdict(list)
    for page in translation_request.pages.all():
        instance = page.source_revision.as_page_object()

        segments = extract_segments(instance)

        # Filter out templates
        text_segments = [segment for segment in segments if isinstance(segment, SegmentValue)]

        for segment in text_segments:
            messages[segment.text].append((instance.url_path, segment.path))

    # Build a PO file
    po = polib.POFile()
    po.metadata = {
        'POT-Creation-Date': str(timezone.now()),
        'MIME-Version': '1.0',
        'Content-Type': 'text/plain; charset=utf-8',
    }

    for text, occurances in messages.items():
        po.append(
            polib.POEntry(
                msgid=text,
                msgstr='',  # TODO: Fetch this from translation memory
                occurrences=occurances
            )
        )

    # Write response
    response = HttpResponse(str(po), content_type='text/x-gettext-translation')
    response['Content-Disposition'] = 'attachment; filename=translation-%d.po' % translation_request.id
    return response


class MissingSegmentsException(Exception):
    def __init__(self, instance, num_missing):
        self.instance = instance
        self.num_missing = num_missing

        super().__init__()


@require_POST
def upload(request, translation_request_id):
    translation_request = get_object_or_404(TranslationRequest, id=translation_request_id)

    with tempfile.NamedTemporaryFile() as f:
        f.write(request.FILES['file'].read())
        f.flush()
        po = polib.pofile(f.name)

    translations = {message.msgid: message.msgstr for message in po}

    try:
        with transaction.atomic():
            for page in translation_request.pages.filter(is_completed=False):
                instance = page.source_revision.as_page_object()

                segments = extract_segments(instance)

                text_segments = [segment for segment in segments if isinstance(segment, SegmentValue)]
                template_segments = [segment for segment in segments if isinstance(segment, TemplateValue)]

                translated_segments = template_segments.copy()

                missing_segments = 0
                for segment in text_segments:
                    if segment.text in translations:
                        translated_segments.append(SegmentValue(segment.path, translations[segment.text]))
                    else:
                        missing_segments += 1

                if missing_segments:
                    raise MissingSegmentsException(instance, missing_segments)

                try:
                    translation = instance.get_translation(translation_request.target_locale)
                except instance.__class__.DoesNotExist:
                    translation = instance.copy_for_translation(translation_request.target_locale)

                ingest_segments(instance, translation, translation_request.source_locale, translation_request.target_locale, translated_segments)
                translation.slug = slugify(translation.slug)
                revision = translation.save_revision()

                # Update translation request
                page.is_completed = True
                page.completed_revision = revision
                page.save(update_fields=['is_completed', 'completed_revision'])

    except MissingSegmentsException as e:
        # TODO: Plural
        messages.error(request, "Unable to translate %s. %s missing segments." % (e.instance.get_admin_display_title(), e.num_missing))

    else:
        # TODO: Plural
        messages.success(request, "%d pages successfully translated with PO file" % translation_request.pages.count())

    return redirect('wagtail_localize_workflow_management:detail', translation_request_id)



from django.db.models import Case, Count, Exists, IntegerField, OuterRef, Q, Sum, Value, When

from wagtail_localize.segments import TemplateValue

from .models import Segment, SegmentTranslation, SegmentPageLocation, TemplatePageLocation


def get_translation_progress(page_revision_id, language):
    """
    For the specified page revision, get the current
    translation progress into the specified language.

    Returns two integers:
     - The total number of segments in the revision to translate
     - The number of segments that have been translated into the language
    """
    # Get QuerySet of Segments that need to be translated
    required_segments = Segment.objects.filter(
        id__in=SegmentPageLocation.objects.filter(page_revision_id=page_revision_id).values_list('segment_id')
    )

    # Annotate each Segment with a flag that indicates whether the segment is translated
    # into the specified language
    required_segments = required_segments.annotate(
        is_translated=Exists(
            SegmentTranslation.objects.filter(
                translation_of_id=OuterRef('pk'),
                language=language,
            )
        )
    )

    # Count the total number of segments and the number of translated segments
    aggs = required_segments.annotate(
        is_translated_i=Case(
            When(is_translated=True, then=Value(1)),
            default=Value(0),
            output_field=IntegerField()
        )
    ).aggregate(
        total_segments=Count('pk'),
        translated_segments=Sum('is_translated_i'),
    )

    return aggs['total_segments'], aggs['translated_segments']


def insert_segments(page_revision, language, segments):
    """
    Inserts the list of untranslated segments into translation memory
    """
    for segment in segments:
        if isinstance(segment, TemplateValue):
            TemplatePageLocation.from_template_value(page_revision, segment)
        else:
            SegmentPageLocation.from_segment_value(page_revision, language, segment)



from collections import Counter

from django.utils.html import escape

from .html import extract_html_elements, restore_html_elements


class SegmentValue:
    class HTMLElement:
        """
        Represents the position of an inline element within an HTML segment value.

        These are used to track how inline elements such as text formatting
        and links should be moved in translated versions of a segment.

        The parameters are as follows:

        - start/end are the character offsets in the original text that this element appears
        they may be equal but end must not be less than start.
        For example, to select just the first character, the start and end offsets with be 0,1
        respectively.
        - identifier is a number that is generated from the segment extractor. It must be conserved
        by translation engines as it is used to work out where elements have moved during translation.
        """
        def __init__(self, start, end, identifier, element=None):
            self.start = start
            self.end = end
            self.identifier = identifier
            self.element = element

    def __init__(self, path, text, html_elements=None, order=0):
        self.path = path
        self.order = order
        self.text = text
        self.html_elements = html_elements

    @classmethod
    def from_html(cls, path, html):
        text, elements = extract_html_elements(html)

        html_elements = []
        counter = Counter()
        for start, end, element_type, element_attrs in elements:
            counter[element_type] += 1
            identifier = element_type + str(counter[element_type])

            html_elements.append(cls.HTMLElement(start, end, identifier, (element_type, element_attrs)))

        return cls(path, text, html_elements)

    def with_order(self, order):
        """
        Sets the order of this segment.
        """
        return SegmentValue(self.path, self.text, order=order)

    def wrap(self, base_path):
        """
        Appends a component to the beginning of the path.

        For example:

        >>> s = SegmentValue('field', "The text")
        >>> s.wrap('relation')
        SegmentValue('relation.field', "The text")
        """
        new_path = base_path

        if self.path:
            new_path += '.' + self.path

        return SegmentValue(new_path, self.text, self.html_elements, order=self.order)

    def unwrap(self):
        """
        Pops a component from the beginning of the path. Reversing .wrap().

        For example:

        >>> s = SegmentValue('relation.field', "The text")
        >>> s.unwrap()
        'relation', SegmentValue('field', "The text")
        """
        base_path, *remaining_components = self.path.split('.')
        new_path = '.'.join(remaining_components)
        return base_path, SegmentValue(new_path, self.text, self.html_elements, order=self.order)

    @property
    def html(self):
        if not self.html_elements:
            return escape(self.text)

        return restore_html_elements(self.text, [
            (e.start, e.end, e.element[0], e.element[1])
            for e in self.html_elements
        ])

    @property
    def html_with_ids(self):
        if not self.html_elements:
            return escape(self.text)

        def cat_dict(a, b):
            c = {}
            c.update(a)
            c.update(b)
            return c

        return restore_html_elements(self.text, [
            (e.start, e.end, e.element[0], {'id': e.identifier} if e.element[1] else {})
            for e in self.html_elements
        ])

    def get_html_element_attrs(self):
        """
        Returns a mapping of html element identifiers to their attributes.

        For example, running this on the following segment:

            <b>Foo <a id="a1" href="https://mysite.com">Bar</a></b>

        Would return the following dictionary:

            {
                "a1": {"href": "https://mysite.com"}
            }
        """
        return {
            e.identifier: e.element[1]
            for e in self.html_elements or []
        }

    def replace_html_element_attrs(self, attrs_map):
        """
        Replaces the attributes of the HTML elements in this segment with
        attributes in the provided mapping.

        This is used to overwrite any HTML attribute changes made to
        translated segments so they are guarenteed to not be altered by
        translators.

        For example, running this on the following segment:

            <b>Foo <a id="a1" href="https://badsite.com">Bar</a></b>

        With the following attrs_map:

            {
                "a1": {"href": "https://mysite.com"}
            }

        Would return the following segment:

            <b>Foo <a id="a1" href="https://mysite.com">Bar</a></b>
        """
        for e in html_elements:
            if e.identifier in attrs_map:
                e.element[1] = attrs_map[e.identifier]

    def is_empty(self):
        return self.html in ['', None]

    def __eq__(self, other):
        return isinstance(other, SegmentValue) and self.path == other.path and self.html == other.html

    def __repr__(self):
        return f'<SegmentValue {self.path} "{self.html}">'


class TemplateValue:
    def __init__(self, path, format, template, segment_count, order=0):
        self.path = path
        self.order = order
        self.format = format
        self.template = template
        self.segment_count = segment_count

    def with_order(self, order):
        """
        Sets the order of this segment.
        """
        return TemplateValue(self.path, self.format, self.template, self.segment_count, order=order)

    def wrap(self, base_path):
        """
        Appends a component to the beginning of the path.

        For example:

        >>> s = TemplateValue('field', 'html', "<text position=\"0\">, 1)
        >>> s.wrap('relation')
        TemplateValue('relation.field', 'html', "<text position=\"0\">, 1)
        """
        new_path = base_path

        if self.path:
            new_path += '.' + self.path

        return TemplateValue(new_path, self.format, self.template, self.segment_count, order=self.order)

    def unwrap(self):
        """
        Pops a component from the beginning of the path. Reversing .wrap().

        For example:

        >>> s = TemplateValue('relation.field', 'html', "<text position=\"0\">, 1)
        >>> s.unwrap()
        'relation', TemplateValue('field', 'html', "<text position=\"0\">, 1)
        """
        base_path, *remaining_components = self.path.split('.')
        new_path = '.'.join(remaining_components)
        return base_path, TemplateValue(new_path, self.format, self.template, self.segment_count, order=self.order)

    def is_empty(self):
        return self.template in ['', None]

    def __eq__(self, other):
        return isinstance(other, TemplateValue) and self.path == other.path and self.format == other.format and self.template == other.template and self.segment_count == other.segment_count

    def __repr__(self):
        return f'<TemplateValue {self.path} format:{self.format} {self.segment_count} segments>'
